from __future__ import annotations

from functools import partial
from warnings import warn

from pyiron_contrib.workflow.has_to_dict import HasToDict
from pyiron_contrib.workflow.node import Node
from pyiron_contrib.workflow.util import DotDict


class _NodeAdder:
    """
    This class exists to help with the misdirection required for the syntactic sugar
    that lets us add nodes to the workflow.

    TODO: Give access to pre-built fixed nodes under various domain names
    """
    def __init__(self, workflow: Workflow):
        self._workflow = workflow

    Node = Node

    def __getattribute__(self, key):
        value = super().__getattribute__(key)
        if value == Node:
            return partial(Node, workflow=self._workflow)
        return value

    def __call__(self, node: Node):
        if node.workflow is not None:
            raise ValueError(
                f"This node ({node.label}) already belongs to the workflow "
                f"{node.workflow.label}. Please remove it there before trying to add it"
                f"to this workflow ({self._workflow.label})."
            )

        if node.label in self._workflow.__dir__() and \
                not isinstance(getattr(self._workflow, node.label), Node):
            raise AttributeError(
                f"Cannot add a node with label {node.label}, that is already an "
                f"attribute"
            )

        if self._workflow.strict_naming and node.label in self._workflow.nodes.keys():
            raise AttributeError(
                f"Cannot add a node with label {node.label}, that is already a node."
            )

        # Otherwise, if not strict then iterate on name
        i = 0
        label = node.label
        while label in self._workflow.nodes.keys():
            warn(f"{label} is already a node; appending an index to the label...")
            label = f"{node.label}{i}"
            i += 1
        node.label = label
        # Or this while loop just terminates immediately if the name is unique

        self._workflow.nodes[node.label] = node
        node.workflow = self._workflow


class Workflow(HasToDict):
    """
    Workflows are an abstraction for holding a collection of related nodes.

    Nodes can be added to the workflow at instantiation or with dot-assignment later on.
    They are then accessible either under the `nodes` dot-dictionary, or just directly
    by dot-access on the workflow object itself.

    The workflow guarantees that each node it owns has a unique within the scope of this
    workflow, and that each node instance appears only once.

    Using the `input` and `output` attributes, the workflow gives access to all the
    IO channels among its nodes which are currently unconnected.

    Examples:
        We allow adding nodes to workflows in five equivalent ways:
        >>> from pyiron_contrib.workflow.workflow import Workflow
        >>> from pyiron_contrib.workflow.node import Node
        >>>
        >>> def fnc(x=0): return x + 1
        >>>
        >>> n1 = Node(fnc, "x", label="n1")
        >>>
        >>> wf = Workflow("my_workflow", n1)  # As *args at instantiation
        >>> wf.add(Node(fnc, "x", label="n2"))  # Passing a node to the add caller
        >>> wf.add.Node(fnc, "y", label="n3")  # Instantiating from add
        >>> wf.n4 = Node(fnc, "y", label="whatever_n4_gets_used")
        >>> # By attribute assignment
        >>> Node(fnc, "x", label="n5", workflow=wf)
        >>> # By instantiating the node with a workflow

        By default, the node naming scheme is strict, so if you try to add a node to a
        label that already exists, you will get an error. This behaviour can be changed
        at instantiation with the `strict_naming` kwarg, or afterwards with the
        `(de)activate_strict_naming()` method(s). When deactivated, repeated assignments
        to the same label just get appended with an index:
        >>> wf.deactivate_strict_naming()
        >>> wf.my_node = Node(fnc, "y", x=0)
        >>> wf.my_node = Node(fnc, "y", x=1)
        >>> wf.my_node = Node(fnc, "y", x=2)
        >>> print(wf.my_node.inputs.x, wf.my_node0.inputs.x, wf.my_node1.inputs.x)
        0, 1, 2

        We can also use pre-built nodes, e.g.
        >>> from pyiron_contrib.workflow import nodes
        >>>
        >>> wf = Workflow("with_prebuilt")
        >>>
        >>> wf.structure = nodes.BulkStructure(repeat=3, cubic=True, element="Al")
        >>> wf.engine = nodes.Lammps(structure=wf.structure.outputs.structure)
        >>> wf.calc = nodes.CalcMD(job=wf.engine.outputs.job)
        >>> wf.plot = nodes.Scatter(
        ...     x=wf.calc.outputs.steps,
        ...     y=wf.calc.outputs.temperature
        ... )


    TODO: Workflows can be serialized.

    TODO: Once you're satisfied with how a workflow is structured, you can export it
        as a macro node for use in other workflows. (Maybe we should allow for nested
        workflows without exporting to a node? I was concerned then what happens to the
        nesting abstraction if, instead of accessing IO through the workflow's IO flags,
        a user manually connects IO from individual nodes from two different, nested or
        sibling workflows when those connections were _previously internal to their own
        workflow_. This seems very unsafe. Maybe there is something like a lock we can
        apply that falls short of a full export, but still guarantees the internal
        integrity of workflows when they're used somewhere else?
    """
    def __init__(self, label: str, *nodes: Node, strict_naming=True):
        self.__dict__['label'] = label
        self.__dict__['nodes'] = DotDict()
        self.__dict__['add'] = _NodeAdder(self)
        self.__dict__['strict_naming'] = strict_naming
        # We directly assign using __dict__ because we override the setattr later

        for node in nodes:
            self.add(node)

    def activate_strict_naming(self):
        self.__dict__['strict_naming'] = True

    def deactivate_strict_naming(self):
        self.__dict__['strict_naming'] = False

    def remove(self, node: Node | str):
        if isinstance(node, Node):
            node.workflow = None
            node.disconnect()
            del self.nodes[node.label]
        else:
            del self.nodes[node]

    def __setattr__(self, label: str, node: Node):
        if label in self.__dir__() and not isinstance(getattr(self, label), Node):
            raise AttributeError(
                f"{label} is already an attribute of {self.label} and cannot be "
                f"reassigned. If this is a node, you can remove the existing node "
                f"first to free the namespace."
            )
        elif not isinstance(node, Node):
            raise TypeError(f"Can only assign nodes, but got {type(node)}")
        else:
            if node.label != label:
                warn(
                    f"Reassigning the node {node.label} to the label {label} when "
                    f"adding it to the workflow {self.label}."
                )
            old_label = node.label
            old_workflow = node.workflow
            try:
                node.label = label
                self.add(node)
            except:
                node.label = old_label
                node.workflow = old_workflow
                raise

    def __getattr__(self, key):
        return self.nodes[key]

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __iter__(self):
        return self.nodes.values().__iter__()

    def __len__(self):
        return len(self.nodes)

    @property
    def input(self):
        return DotDict(
            {
                f"{node.label}_{channel.label}": channel
                for node in self.nodes.values()
                for channel in node.inputs
                if not channel.connected
            }
        )

    @property
    def output(self):
        return DotDict(
            {
                f"{node.label}_{channel.label}": channel
                for node in self.nodes.values()
                for channel in node.outputs
                if not channel.connected
            }
        )

    def to_dict(self):
        return {
            "label": self.label,
            "nodes": {n.label: n.to_dict() for n in self.nodes.values()},
        }

    def to_node(self):
        """
        Export the workflow to a macro node, with the currently exposed IO mapped to
        new IO channels, and the workflow mapped into the node_function.
        """
        raise NotImplementedError

    # (De)serialization is necessary throughout these classes, but not implemented here
    def serialize(self):
        raise NotImplementedError

    def deserialize(self, source):
        raise NotImplementedError

    def update(self):
        for node in self.nodes.values():
            if node.outputs.connected and not node.inputs.connected:
                node.update()

    def run(self):
        # Maybe we need this if workflows can be used as nodes?
        raise NotImplementedError

    def __dir__(self):
        return set(super().__dir__() + list(self.nodes.keys()))



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import os
import shutil
from pyiron_atomistics.atomistics.job.atomistic import AtomisticGenericJob
from pyiron_contrib.atomistics.mlip.mlip import write_cfg, read_cgfs
from pyiron_base import GenericParameters

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - " \
                "Computational Materials Design (CM) Department"
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"


class MlipJob(AtomisticGenericJob):
    def __init__(self, project, job_name):
        super(MlipJob, self).__init__(project, job_name)
        self.__name__ = "MlipJob"
        self.__version__ = None
        self._executable_activate()
        self.input = MlipParameter()

    @property
    def potential(self):
        return self.input['mlip:load-from']

    @potential.setter
    def potential(self, potential_filename):
        self.input['mlip:load-from'] = potential_filename

    def set_input_to_read_only(self):
        """
        This function enforces read-only mode for the input classes, but it has to be implement in the individual
        classes.
        """
        super(MlipJob, self).set_input_to_read_only()
        self.input.read_only = True

    def to_hdf(self, hdf=None, group_name=None):
        super(MlipJob, self).to_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("input") as hdf5_input:
            self.input.to_hdf(hdf5_input)

    def from_hdf(self, hdf=None, group_name=None):
        super(MlipJob, self).from_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("input") as hdf5_input:
            self.input.from_hdf(hdf5_input)

    def write_input(self):
        write_cfg(file_name=os.path.join(self.working_directory, 'structure.cfg'),
                  indices_lst=[self.structure.indices],
                  position_lst=[self.structure.positions],
                  cell_lst=[self.structure.cell],
                  forces_lst=None, energy_lst=None, track_lst=None, stress_lst=None)
        shutil.copyfile(self.input['mlip:load-from'], os.path.join(self.working_directory, 'potential.mtp'))
        self.input['mlip:load-from'] = 'potential.mtp'
        self.input.write_file(file_name="mlip.ini", cwd=self.working_directory)

    def collect_output(self):
        file_name = os.path.join(self.working_directory, 'structurebyMTP.cfg')
        cell, positions, forces, stress, energy, indicies, grades, jobids, timesteps = read_cgfs(file_name=file_name)
        with self.project_hdf5.open('output') as hdf5_output:
            hdf5_output['forces'] = forces
            hdf5_output['energy_tot'] = energy
            hdf5_output['pressures'] = stress
            hdf5_output['cells'] = cell
            hdf5_output['positions'] = positions
            hdf5_output['indicies'] = indicies


class MlipParameter(GenericParameters):
    def __init__(self, separator_char=' ', comment_char='#', table_name="mlip_inp"):
        super(MlipParameter, self).__init__(separator_char=separator_char, comment_char=comment_char,
                                            table_name=table_name)
        self._bool_dict = {True: "TRUE", False: "FALSE"}

    def load_default(self, file_content=None):
        if file_content is None:
            file_content = '''\
abinitio void
mlip mtpr
mlip:load-from auto
calculate-efs TRUE
write-cfgs structurebyMTP.cfg
'''
        self.load_string(file_content)



import xml.etree.ElementTree as ET
import sys
import os
import threading
from xml.dom import minidom
import time
import select
import ctypes


def write_pretty_xml(elem, filename):
    """
    The files directly written with xml.etree.ElemenTrees contain no newlines and indents.
    This is a helper function to write them in a human readable way.

    Args:
        elem (ElementTree): xml.etree.ElementTree ElementTree object to be written
        filename (string): name for the xml file
    """
    ugly = ET.tostring(elem, "utf-8")
    reparsed = minidom.parseString(ugly)
    reparsed = reparsed.toprettyxml(indent="    ")
    with open(filename, "w") as f:
        f.write(reparsed)
    return


"""
This was supposed to catch the output of atomicrex when running it using its python interface.
It never worked completely but maybe this will be useful for some other code

class OutputCatcher():
    libc = ctypes.CDLL(None)
    c_stdout = ctypes.c_void_p.in_dll(libc, 'stdout')

    def __init__(self, filename, threaded=False):
        sys.stdout.flush()
        self.filename = filename
        #self.pipe_in, self.pipe_out = os.pipe()
        ## Maybe use tempfile instead?
        self.file_out = os.open(self.filename, os.O_WRONLY|os.O_TRUNC|os.O_CREAT)
        self.threaded = threaded
        self.libc.fflush(self.c_stdout)

    def __enter__(self):
        self.stdout_old = os.dup(1)
        self.stdout_new = os.dup(1)
        os.dup2(self.file_out, 1)
        sys.stdout = os.fdopen(self.stdout_new, 'w')
        if self.threaded:
            self.worker = threading.Thread(target=self.read_output)
            self.worker.start()
            time.sleep(0.01)
 
    def __exit__(self, exc_type, exc_val, exc_tb):
        os.close()
        if self.threaded:
            self.worker.join()
        else:
            self.read_output()
        self.libc.fflush(self.c_stdout)
        os.dup2(self.stdout_old, 1)
        os.close(self.stdout_old)
        os.close(self.file_out)

    def read_output(self):
        r, _, _ = select.select([self.stdout_new], [], [], 0)
        while bool(r):
            char = os.read(self.stdout_new, 1024)
            self.file_out.write(char)
            r, _, _ = select.select([self.stdout_new], [], [], 0)
"""



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import os, posixpath
import xml.etree.ElementTree as ET

import numpy as np
from numpy import ndarray
from pyiron_base import state, DataContainer, FlattenedStorage
from pyiron_atomistics import Atoms, ase_to_pyiron

from pyiron_atomistics.atomistics.structure.structurestorage import StructureStorage
from pyiron_contrib.atomistics.atomistics.job.trainingcontainer import (
    TrainingPlots,
    TrainingContainer,
    TrainingStorage,
)
from pyiron_contrib.atomistics.atomicrex.fit_properties import (
    ARFitPropertyList,
    ARFitProperty,
)
from pyiron_contrib.atomistics.atomicrex.utility_functions import write_pretty_xml
from pyiron_contrib.atomistics.atomicrex.fit_properties import (
    FlattenedARScalarProperty,
    FlattenedARVectorProperty,
)

try:
    import atomicrex
except ImportError:
    pass


class ARStructureContainer:
    __version__ = "0.3.0"
    __hdf_version__ = "0.3.0"

    def __init__(self, num_atoms=0, num_structures=0):
        self.fit_properties = DataContainer(table_name="fit_properties")
        self._structures = StructureStorage(
            num_atoms=num_atoms, num_structures=num_structures
        )
        self._predefined_storage = DataContainer(table_name="predefined_structures")
        self._structures.add_array("fit", dtype=bool, per="chunk", fill=True)
        self._structures.add_array("clamp", dtype=bool, per="chunk", fill=True)
        self._structures.add_array("predefined", dtype=bool, per="chunk", fill=False)
        self._structures.add_array("relative_weight", per="chunk", fill=1.0)
        self.structure_file_path = None
        try:
            self._interactive_library = atomicrex.Job()
            for path in state.settings.resource_paths:
                path = posixpath.join(
                    state.settings.resource_paths[0], "atomicrex/util/main.xml"
                )
                if os.path.isfile(path):
                    self._interactive_library.parse_input_file(path)
                    break
        except:
            pass
        self._prepared_plotting = False

    def add_structure(
        self,
        structure,
        identifier,
        fit=True,
        relative_weight=1,
        clamp=True,
    ):
        if "/" in identifier:
            raise ValueError(
                "Structure identifiers must not contain '/'. "
                "Use .structure_file_path to use existing POSCAR files"
            )

        self._structures.add_structure(structure, identifier)
        self._structures._per_chunk_arrays["fit"][
            self._structures.prev_chunk_index
        ] = fit
        self._structures._per_chunk_arrays["relative_weight"][
            self._structures.prev_chunk_index
        ] = relative_weight
        self._structures._per_chunk_arrays["clamp"][
            self._structures.prev_chunk_index
        ] = clamp
        self._structures._per_chunk_arrays["predefined"][
            self._structures.prev_chunk_index
        ] = False

    def add_predefined_structure(
        self,
        identifier,
        lattice,
        lattice_parameter,
        atom_type_A,
        ca_ratio=None,
        atom_type_B=None,
        fit=True,
        relative_weight=1,
        clamp=True,
    ):
        data = {
            "alat": lattice_parameter,
        }
        if atom_type_B is None:
            data["type"] = atom_type_A
        else:
            data["type_A"] = atom_type_A
            data["type_B"] = atom_type_B
        if ca_ratio is not None:
            data["ca_ratio"] = ca_ratio
        struct = self._interactive_library.add_library_structure(
            identifier, lattice, data
        )
        struct = ase_to_pyiron(struct.get_atoms(self._interactive_library))

        self._structures.add_structure(struct, identifier)
        self._structures._per_chunk_arrays["fit"][
            self._structures.prev_chunk_index
        ] = fit
        self._structures._per_chunk_arrays["relative_weight"][
            self._structures.prev_chunk_index
        ] = relative_weight
        self._structures._per_chunk_arrays["clamp"][
            self._structures.prev_chunk_index
        ] = clamp
        self._structures._per_chunk_arrays["predefined"][
            self._structures.prev_chunk_index
        ] = True

        storage = DataContainer(table_name=identifier)
        storage["lattice"] = lattice
        storage["lattice_parameter"] = lattice_parameter
        storage["ca_ratio"] = ca_ratio
        storage["atom_type_A"] = atom_type_A
        storage["atom_type_B"] = atom_type_B
        self._predefined_storage[identifier] = storage

    def add_scalar_fit_property(
        self,
        prop="atomic-energy",
        target_val=np.nan,
        fit=True,
        relax=False,
        relative_weight=1,
        residual_style=0,
        output=True,
        tolerance=np.nan,
        min_val=np.nan,
        max_val=np.nan,
    ):
        try:
            flat = self.fit_properties[prop]
        except KeyError:
            self.fit_properties[prop] = FlattenedARScalarProperty(
                num_chunks=self._structures.num_chunks,
                num_elements=self._structures.num_elements,
            )
            flat = self.fit_properties[prop]
            if prop in ["lattice-parameter", "ca-ratio"]:
                flat.add_array("min_val", per="chunk")
                flat.add_array("max_val", per="chunk")
        try:
            flat._per_chunk_arrays["target_val"][
                self._structures.prev_chunk_index
            ] = target_val
        except (IndexError, ValueError):
            for v in self.fit_properties.values():
                v._resize_chunks(self._structures._num_chunks_alloc)
                v._resize_elements(self._structures._num_elements_alloc)
            flat._per_chunk_arrays["target_val"][
                self._structures.prev_chunk_index
            ] = target_val
        flat._per_chunk_arrays["fit"][self._structures.prev_chunk_index] = fit
        flat._per_chunk_arrays["relax"][self._structures.prev_chunk_index] = relax
        flat._per_chunk_arrays["relative_weight"][
            self._structures.prev_chunk_index
        ] = relative_weight
        flat._per_chunk_arrays["residual_style"][
            self._structures.prev_chunk_index
        ] = residual_style
        flat._per_chunk_arrays["output"][self._structures.prev_chunk_index] = output
        flat._per_chunk_arrays["tolerance"][
            self._structures.prev_chunk_index
        ] = tolerance
        if prop in ["lattice-parameter", "ca-ratio"]:
            flat._per_chunk_arrays["min_val"][
                self._structures.prev_chunk_index
            ] = min_val
            flat._per_chunk_arrays["max_val"][
                self._structures.prev_chunk_index
            ] = max_val

    def add_vector_fit_property(
        self,
        prop="atomic-forces",
        target_val=None,
        fit=True,
        relax=False,
        relative_weight=1,
        residual_style=0,
        tolerance=np.nan,
        output=True,
    ):
        try:
            flat = self.fit_properties[prop]
        except KeyError:
            self.fit_properties[prop] = FlattenedARVectorProperty(
                num_chunks=self._structures.num_chunks,
                num_elements=self._structures.num_elements,
            )
            flat = self.fit_properties[prop]
        try:
            flat._per_chunk_arrays["fit"][self._structures.prev_chunk_index] = fit
        except (IndexError, ValueError):
            for v in self.fit_properties.values():
                v._resize_chunks(self._structures._num_chunks_alloc)
            flat._per_chunk_arrays["fit"][self._structures.prev_chunk_index] = fit

        if target_val is not None:
            try:
                flat._per_element_arrays["target_val"][
                    self._structures.prev_element_index : self._structures.current_element_index
                ] = target_val
            except (IndexError, ValueError):
                for v in self.fit_properties.values():
                    v._resize_elements(self._structures._num_elements_alloc)
                flat._per_element_arrays["target_val"][
                    self._structures.prev_element_index : self._structures.current_element_index
                ] = target_val
        flat._per_chunk_arrays["relax"][self._structures.prev_chunk_index] = relax
        flat._per_chunk_arrays["relative_weight"][
            self._structures.prev_chunk_index
        ] = relative_weight
        flat._per_chunk_arrays["residual_style"][
            self._structures.prev_chunk_index
        ] = residual_style
        flat._per_chunk_arrays["output"][self._structures.prev_chunk_index] = output
        flat._per_chunk_arrays["tolerance"][
            self._structures.prev_chunk_index
        ] = tolerance

    def _get_per_structure_index(self, identifier):
        """
        Takes an identifier or an ndarray of indentifiers and returns the corresponding indices in the structure/scalar_fit_properties arrays.
        Args:
            identifiers ([type]): [description]

        Returns:
            [ndarray]: indices corresponding to identifiers in per structure arrays
        """
        if not isinstance(identifier, ndarray):
            identifier = np.array(identifier)
        indices = np.flatnonzero(
            np.isin(
                self._structures._per_chunk_arrays["identifier"],
                identifier,
                assume_unique=True,
            )
        )
        # This is some sorting magic that could lead to strange errors
        # Look here if something goes wrong with the ordering
        ids_stored = np.array(self._structures._per_chunk_arrays["identifier"][indices])
        sorter = ids_stored.argsort()[identifier.argsort()]
        return indices[sorter]

    def get_scalar_property(self, prop, identifier, final=True):
        """
        Returns final or target value of a scalar property based on the identifier used when adding a structure
        If identifier is an array of identifiers the return value is an array of values.

        Args:
            prop (str): property string as used when adding the property
            identifier (str or array(str)): identifier as used when adding the structure
            final (bool, optional): Whether to return the final or target value. Defaults to True.

        Returns:
            [type]: [description]
        """
        index = self._get_per_structure_index(identifier)
        if final:
            return self.fit_properties[prop]._per_chunk_arrays["final_val"][index]
        else:
            return self.fit_properties[prop]._per_chunk_arrays["target_val"][index]

    def get_vector_property(self, prop, identifier, final=True):
        """
        Returns final or target value of a vector property based on the identifier used when adding a structure.
        Currently only allows to return a vector property for a single structure, not for multiple ones like the
        get_scalar_property function.

        Args:
            prop (str): property string as used when adding the property
            identifier (str): identifier as used when adding the structure
            final (bool, optional): Whether to return the final or target value. Defaults to True.

        Returns:
            [type]: [description]
        """
        if not isinstance(identifier, str):
            raise NotImplementedError(
                "Can only look up properties for single identifiers currently"
            )
        index = self._get_per_structure_index(identifier)[0]
        slc = self._structures._get_per_element_slice(index)
        if final:
            return self.fit_properties[prop]._per_element_arrays["final_val"][slc]
        else:
            return self.fit_properties[prop]._per_element_arrays["target_val"][slc]

    def _sync(self):
        for flat in self.fit_properties.values():
            flat.num_elements = self._structures.num_elements
            flat.num_chunks = self._structures.num_chunks

    def _shrink(self):
        self._resize_all(
            num_chunks=self._structures.num_chunks,
            num_elements=self._structures.num_elements,
        )

    def _resize_all(self, num_chunks, num_elements):
        self._structures.num_elements = num_elements
        self._structures.num_chunks = num_chunks
        self._structures._resize_elements(self._structures.num_elements)
        self._structures._resize_chunks(self._structures.num_chunks)
        for flat in self.fit_properties.values():
            flat.num_elements = num_elements
            flat.num_chunks = num_chunks
            flat._resize_elements(num_elements)
            flat._resize_chunks(num_chunks)

    def _type_to_hdf(self, hdf):
        """
        Internal helper function to save type and version in hdf root

        Args:
            hdf (ProjectHDFio): HDF5 group object
        """
        hdf["NAME"] = self.__class__.__name__
        hdf["TYPE"] = str(type(self))
        hdf["VERSION"] = self.__version__
        hdf["HDF_VERSION"] = self.__hdf_version__

    def to_hdf(self, hdf, group_name="structures"):
        self._shrink()
        with hdf.open(group_name) as h:
            self._type_to_hdf(h)
            self._structures.to_hdf(hdf=h)
            self.fit_properties.to_hdf(
                hdf=h,
            )
            self._predefined_storage.to_hdf(hdf=h)
            h["structure_file_path"] = self.structure_file_path

    def from_hdf(self, hdf, group_name="structures"):
        with hdf.open(group_name) as h:
            version = h.get("HDF_VERSION", "0.1.0")
            # Compatibility Old and new StructureStorage
            if version == "0.1.0":
                num_structures = h["flattened_structures/num_structures"]
                num_atoms = h["flattened_structures/num_atoms"]
                group_name_2 = "flattened_structures"
            else:
                try:
                    num_structures = h["structures/num_chunks"]
                    num_atoms = h["structures/num_elements"]
                except:
                    num_structures = h["structures/num_structures"]
                    num_atoms = h["structures/num_atoms"]
                group_name_2 = "structures"

            self._resize_all(num_chunks=num_structures, num_elements=num_atoms)
            self._structures.from_hdf(hdf=h, group_name=group_name_2)
            self.structure_file_path = h["structure_file_path"]

            if version == "0.3.0":
                self.fit_properties.from_hdf(hdf=h)
                self._predefined_storage.from_hdf(hdf=h)
            else:
                with h.open("fit_properties") as g:
                    for k in g.list_groups():
                        if k == "atomic-forces":
                            self.fit_properties[k] = FlattenedARVectorProperty(
                                num_chunks=num_structures, num_elements=num_atoms
                            )
                        else:
                            self.fit_properties[k] = FlattenedARScalarProperty(
                                num_chunks=num_structures, num_elements=num_atoms
                            )
                        self.fit_properties[k].from_hdf(hdf=g, group_name=k)

            if version < "0.3.0":
                self._structures._per_chunk_arrays["clamp"] = h["clamp"]
                self._structures._per_chunk_arrays["fit"] = h["fit"]
                self._structures._per_chunk_arrays["relative_weight"] = h[
                    "relative_weight"
                ]

    def _check_identifiers(self):
        identifiers = self._structures.get_array("identifier")
        if not np.all(np.char.find(identifiers, "/")==-1):
            raise ValueError(
                "Structure identifiers must not contain '/'. "
                "Use .structure_file_path to use existing POSCAR files"
            )

    def write_xml_file(self, directory, name="structures.xml"):
        """
        Internal helper function that writes an atomicrex style
        xml file containg all structures.

        Args:
            directory (string): Working directory.
            name (str, optional): . Defaults to "structures.xml".
        """
        self._check_identifiers()
        self._shrink()
        root = ET.Element("group")
        if self.structure_file_path is None and "atomic-forces" in self.fit_properties:
            # write POSCARs
            for i in range(self._structures.num_chunks):
                vec_start = self._structures.start_index[i]
                vec_end = self._structures.start_index[i] + self._structures.length[i]
                forces = self.fit_properties["atomic-forces"]._per_element_arrays[
                    "target_val"
                ][vec_start:vec_end]
                if not self._structures._per_chunk_arrays["predefined"][i]:
                    write_modified_poscar(
                        identifier=self._structures.identifier[i],
                        forces=forces,
                        positions=self._structures.positions[vec_start:vec_end],
                        symbols=self._structures.symbols[vec_start:vec_end],
                        cell=self._structures.cell[i],
                        directory=directory,
                    )
                # Maybe implement a check for forces when setting up a predefined structure?
                # else:
                #    if not np.all(np.isnan(forces)):

        # write xml
        for i in range(self._structures.num_chunks):
            fit_properties_xml = ET.Element("properties")
            if not self._structures._per_chunk_arrays["predefined"][i]:
                for prop, flat_prop in self.fit_properties.items():
                    if not prop in ("lattice-parameter", "ca-ratio"):
                        fit_properties_xml.append(flat_prop.to_xml_element(i, prop))
                struct_xml = structure_meta_xml(
                    identifier=self._structures.identifier[i],
                    relative_weight=self._structures._per_chunk_arrays[
                        "relative_weight"
                    ][i],
                    clamp=self._structures._per_chunk_arrays["clamp"][i],
                    fit_properties=fit_properties_xml,
                    struct_file_path=self.structure_file_path,
                    fit=self._structures._per_chunk_arrays["fit"][i],
                )
            else:
                for prop, flat_prop in self.fit_properties.items():
                    fit_properties_xml.append(flat_prop.to_xml_element(i, prop))
                data = self._predefined_storage[self._structures.identifier[i]]
                struct_xml = predefined_structure_xml(
                    identifier=self._structures.identifier[i],
                    lattice=data["lattice"],
                    lattice_param=data["lattice_parameter"],
                    ca_ratio=data["ca_ratio"],
                    atom_type_A=data["atom_type_A"],
                    atom_type_B=data["atom_type_B"],
                    relative_weight=self._structures._per_chunk_arrays[
                        "relative_weight"
                    ][i],
                    clamp=self._structures._per_chunk_arrays["clamp"][i],
                    fit_properties=fit_properties_xml,
                    fit=self._structures._per_chunk_arrays["fit"][i],
                )
            root.append(struct_xml)
        filename = posixpath.join(directory, name)
        write_pretty_xml(root, filename)

    def _parse_final_properties(self, struct_lines):
        """
        Internal function that parses the values of fitted properties
        calculated with the final iteration of the fitted potential.

        Args:
            struct_lines (list[str]): lines from atomicrex output that contain structure information.
        """
        force_vec_triggered = False
        for l in struct_lines:
            l = l.strip()

            if force_vec_triggered:
                if l.startswith("atomic-forces:"):
                    l = l.split()
                    index = int(l[1])
                    final_forces[index, 0] = float(l[2].lstrip("(").rstrip(","))
                    final_forces[index, 1] = float(l[3].rstrip(","))
                    final_forces[index, 2] = float(l[4].rstrip(")"))
                else:
                    force_vec_triggered = False
                    start_index = self._structures.start_index[s_index]
                    self.fit_properties["atomic-forces"]._per_element_arrays[
                        "final_val"
                    ][start_index : start_index + len_struct] = final_forces

            # This has to be if and not else because it has to run in the same iteration. Empty lines get skipped.
            if not force_vec_triggered and l:
                if l.startswith("Structure"):
                    s_id = l.split("'")[1]
                    s_index = np.nonzero(self._structures.identifier == s_id)[0][0]

                else:
                    if not l.startswith("atomic-forces avg/max:"):
                        l = l.split()
                        prop, f_val = l[0].rstrip(":"), float(l[1])
                        if prop in self.fit_properties.keys():
                            self.fit_properties[prop]._per_chunk_arrays["final_val"][
                                s_index
                            ] = f_val
                    else:
                        force_vec_triggered = True
                        len_struct = self._structures._per_chunk_arrays["length"][
                            s_index
                        ]
                        final_forces = np.empty((len_struct, 3))

    @property
    def plot(self):
        """
        :class:`.TrainingPlots`: plotting interface
        """
        return TrainingPlots(self._structures)

    def prepare_plotting(self, final_values=False):
        val_str = "target_val"
        if final_values:
            val_str = "final_val"
        self._structures._per_chunk_arrays["energy"] = (
            self.fit_properties["atomic-energy"]._per_chunk_arrays[val_str]
            * self._structures.length
        )

    #### PotentialFit methods
    def add_training_data(self, container: TrainingContainer) -> None:
        storage = container._container
        atomic_energy_storage = FlattenedARScalarProperty(
            num_chunks=storage.num_chunks, num_elements=storage.num_elements
        )
        atomic_energy_storage.num_chunks = storage.num_chunks
        atomic_energy_storage.num_elements = storage.num_elements
        atomic_energy_storage._per_chunk_arrays["fit"][0 : storage.num_chunks] = True
        atomic_energy_storage._per_chunk_arrays["tolerance"][
            0 : storage.num_chunks
        ] = 0.001
        atomic_energy_storage._per_chunk_arrays["target_val"] = (
            storage._per_chunk_arrays["energy"][0 : storage.num_chunks]
            / storage.length[0 : storage.num_chunks]
        )

        atomic_forces_storage = FlattenedARVectorProperty(
            num_chunks=storage.num_chunks, num_elements=storage.num_elements
        )
        atomic_forces_storage.num_chunks = storage.num_chunks
        atomic_forces_storage.num_elements = storage.num_elements
        atomic_forces_storage._per_chunk_arrays["fit"][0 : storage.num_chunks] = True
        atomic_forces_storage._per_chunk_arrays["tolerance"][
            0 : storage.num_chunks
        ] = 0.01
        atomic_forces_storage._per_element_arrays[
            "target_val"
        ] = storage._per_element_arrays["forces"][0 : storage.num_elements]

        if "atomic-energy" not in self.fit_properties:
            self.fit_properties["atomic-energy"] = FlattenedARScalarProperty(
                num_chunks=self._structures.num_chunks,
                num_elements=self._structures.num_elements,
            )
        if "atomic-forces" not in self.fit_properties:
            self.fit_properties["atomic-forces"] = FlattenedARVectorProperty(
                num_chunks=self._structures.num_chunks,
                num_elements=self._structures.num_elements,
            )
        self._sync()

        self._structures.extend(storage)
        self.fit_properties["atomic-energy"].extend(atomic_energy_storage)
        self.fit_properties["atomic-forces"].extend(atomic_forces_storage)

    def _to_TrainingStorage(self, final: bool = False) -> TrainingStorage:
        self._shrink()
        storage = TrainingStorage()
        storage.extend(self._structures)
        if final:
            val_str = "final_val"
        else:
            val_str = "target_val"
        storage._per_chunk_arrays["energy"][0 : storage.num_chunks] = (
            self.fit_properties["atomic-energy"][val_str]
            * self._structures._per_chunk_arrays["length"]
        )
        storage._per_element_arrays["forces"][
            0 : storage.num_elements
        ] = self.fit_properties["atomic-forces"][val_str]
        return storage

    def get_training_data(self) -> TrainingStorage:
        return self._to_TrainingStorage(final=False)

    def get_predicted_data(self) -> FlattenedStorage:
        return self._to_TrainingStorage(final=True)

    @property
    def training_data(self):
        return self.get_training_data()

    @property
    def predicted_data(self):
        return self.get_predicted_data()


### This is probably useless like this in most cases because forces can't be passed.
def user_structure_to_xml_element(structure):
    """
    Converts an ase/pyiron atoms object to an atomicrex xml element
    Right now forces can't be passed in the xml file, so this is not really helpful.
    Args:
        structure (Atoms): ase or pyiron Atoms

    Returns:
        (ET.Element, ET.Element): atomicrex structure xml
    """
    pbc = ET.Element("pbc")
    pbc.set("x", f"{structure.pbc[0]}".lower())
    pbc.set("y", f"{structure.pbc[1]}".lower())
    pbc.set("z", f"{structure.pbc[2]}".lower())

    c = structure.cell
    cell = ET.Element("cell")
    for i in range(3):
        a = ET.SubElement(cell, f"a{i+1}")
        a.set("x", f"{c[i][0]}")
        a.set("y", f"{c[i][1]}")
        a.set("z", f"{c[i][2]}")

    atoms = ET.SubElement(cell, "atoms")
    for at in structure:
        a = ET.SubElement(atoms, "atom")
        a.set("type", at.symbol)
        a.set("x", f"{at.a}")
        a.set("y", f"{at.b}")
        a.set("z", f"{at.c}")
        a.set("reduced", "false")

    return pbc, cell


def write_modified_poscar(
    identifier,
    forces,
    directory,
    structure=None,
    positions=None,
    cell=None,
    symbols=None,
):
    """
    Internal function. Writes a pyiron structure
    and corresponding forces in a modified POSCAR file.
    Either provide a structure instance or positions and symbols.
    Args:
        identifier (str): Unique identifier used for the filename.
        structure (Atoms): ase or pyiron atoms object.
        positions (np.array): atomic positions
        symbols (np.array): chemical symbols
        forces (array or list[list]): atomic forces. Must be in same order as positions.
        directory (str): Working directory.
    """
    filename = posixpath.join(directory, f"POSCAR_{identifier}")
    with open(filename, "w") as f:
        # Elements as system name
        # Also check if the symbols are disordered
        # to prevent errors resulting from poscar format
        # not having indices or element for each atom.
        elements = []
        counter = 0
        last_elem = "XX"
        if symbols is None:
            symbols = np.array(structure.symbols)
        for elem in symbols:
            if not elem in elements:
                elements.append(elem)
            if not last_elem == elem:
                counter += 1
                last_elem = elem
        if counter != len(elements):
            raise ValueError(
                "Structures with disordered elements are not supported right now\n"
                "They can be sorted together with target forces using numpy.argsort"
            )
            ## maybe this can be fixed returning an argsort array to sort the forces.
        f.write(f"{elements}\n")

        # Cell metric
        f.write("1.0\n")
        if cell is None:
            cell = structure.cell
        for vec in cell:
            for v in vec:
                f.write(f"{v} ")
            f.write("\n")

        # Element names
        for elem in elements:
            f.write(f"{elem} ")
        f.write("\n")

        # Number of elements per type
        for elem in elements:
            n = len(symbols[symbols == elem])
            f.write(f"{n}    ")
        f.write("\n")

        # Scaled coordinates
        if positions is None:
            positions = structure.positions
        f.write("Cartesian\n")

        # Coordinates and forces
        for pos, force in zip(positions, forces):
            for r in pos:
                f.write((f"{r} "))
            f.write("   ")
            for r in force:
                f.write((f"{r} "))
            f.write("\n")


def structure_meta_xml(
    identifier,
    relative_weight,
    clamp,  ## Not sure if and how this combines with relax in the ARFitParameter sets
    fit_properties,
    struct_file_path,
    fit,
    mod_poscar=True,
):
    """
    Internal function. Creates xml element with
    scalar properties, weight and reference to POSCAR
    containg structure and forces.

    Args:
        identifier (str): Unique identifier.
        relative_weight (float): weight in objective function
        clamp (bool): clamp the structure (no relaxation)
        mod_poscar (bool, optional): Combine with poscar file. Defaults to True.
        struct_file_path (string): path to POSCAR file
    Raises:
        NotImplementedError: mod_poscar has to be True, because forces can't
        be provided in xml format in atomicrex yet.

    Returns:
        [ElementTree xml element]: atomicrex structure xml element.
    """

    struct_xml = ET.Element("user-structure")
    struct_xml.set("id", f"{identifier}")
    struct_xml.set("relative-weight", f"{relative_weight}")
    if fit:
        struct_xml.set("fit", "true")
    else:
        struct_xml.set("fit", "false")

    if mod_poscar:
        poscar_file = ET.SubElement(struct_xml, "poscar-file")
        if struct_file_path is None:
            poscar_file.text = f"POSCAR_{identifier}"
        else:
            poscar_file.text = f"{struct_file_path}POSCAR_{identifier}"
    else:
        #    struct_xml.extend(structure_to_xml_element(structure))
        raise NotImplementedError(
            "Only writing structure meta information, \n"
            "not structure data for now because this is not implemented in atomicrex"
        )

    if not clamp:
        relax_dof = ET.SubElement(struct_xml, "relax-dof")
        atom_coordinates = ET.SubElement(relax_dof, "atom-coordinates")

    if isinstance(fit_properties, ARFitPropertyList):
        properties = ET.SubElement(struct_xml, "properties")
        for prop in fit_properties.values():
            properties.append(prop.to_xml_element())
    else:
        struct_xml.append(fit_properties)
    return struct_xml


def predefined_structure_xml(
    identifier,
    lattice,
    lattice_param,
    ca_ratio,
    atom_type_A,
    atom_type_B,
    relative_weight,
    clamp,  ## Not sure if and how this combines with relax in the ARFitParameter sets
    fit_properties,
    fit,
):
    """
    Internal function. Creates xml element for predefined structure.
    Args:
        identifier (str): Unique identifier.
        relative_weight (float): weight in objective function
        clamp (bool): clamp the structure (no relaxation)
        fit (bool): whether to fit the structure

    Returns:
        [ElementTree xml element]: atomicrex predefined structure xml element.
    """

    struct_xml = ET.Element(f"{lattice}-lattice")
    struct_xml.set("id", f"{identifier}")
    struct_xml.set("relative-weight", f"{relative_weight}")
    if fit:
        struct_xml.set("fit", "true")
    else:
        struct_xml.set("fit", "false")

    if atom_type_A is None:
        raise ValueError("atom type A has to be given for predefined structures")
    else:
        if atom_type_B is None:
            atomA = ET.SubElement(struct_xml, "atom-type")
        else:
            atomA = ET.SubElement(struct_xml, "atom-type-A")
            atomB = ET.SubElement(struct_xml, "atom-type-B")
            atomB.text = atom_type_B
        atomA.text = atom_type_A

    if lattice_param is None:
        raise ValueError("lattice parameter has to be set for predefined structures")
    else:
        a = ET.SubElement(struct_xml, "lattice-parameter")
        a.text = f"{lattice_param}"

    if ca_ratio is not None:
        ca = ET.SubElement(struct_xml, "ca-ratio")
        ca.text = f"{ca_ratio}"

    if not clamp:
        relax_dof = ET.SubElement(struct_xml, "relax-dof")
        ET.SubElement(relax_dof, "atom-coordinates")

    if isinstance(fit_properties, ARFitPropertyList):
        properties = ET.SubElement(struct_xml, "properties")
        for prop in fit_properties.values():
            properties.append(prop.to_xml_element())
    else:
        struct_xml.append(fit_properties)
    return struct_xml


def sort_structure(structure, forces=None):
    sort_array = get_sort_array(structure)
    if forces is None:
        return structure[sort_array]
    return structure[sort_array], forces[sort_array]


def get_sort_array(structure):
    sort_array = np.argsort(structure.numbers, kind="stable")
    return sort_array



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function
from itertools import islice
import numbers
import numpy as np
import os
import pandas as pd
import posixpath
import scipy.constants
from pyiron_atomistics.atomistics.structure.structurestorage import StructureStorage
from pyiron_base import state, GenericParameters, GenericJob, Executable, FlattenedStorage
from pyiron_atomistics import ase_to_pyiron, Atoms
from pyiron_contrib.atomistics.ml.potentialfit import PotentialFit
from pyiron_contrib.atomistics.mlip.cfgs import savecfgs, loadcfgs, Cfg
from pyiron_contrib.atomistics.mlip.potential import MtpPotential

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - " \
                "Computational Materials Design (CM) Department"
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2017"


gpa_to_ev_ang = 1e22 / scipy.constants.physical_constants['joule-electron volt relationship'][0]

class Mlip(GenericJob, PotentialFit):
    def __init__(self, project, job_name):
        super(Mlip, self).__init__(project, job_name)
        self.__version__ = '0.1.0'
        self.__name__ = "Mlip"
        self._executable_activate()
        self._job_dict = {}
        self.input = MlipParameter()
        self._command_line = CommandLine()
        self._potential = MtpPotential()

    def _executable_activate(self, enforce=False):
        if self._executable is None or enforce:
            if len(self.__module__.split(".")) > 1:
                self._executable = Executable(
                    codename=self.__name__,
                    module=self.__module__.split(".")[-2],
                    path_binary_codes=state.settings.resource_paths,
                )
            else:
                self._executable = Executable(
                    codename=self.__name__, path_binary_codes=state.settings.resource_paths
                )

    @property
    def calculation_dataframe(self):
        job_id_lst, time_step_start_lst, time_step_end_lst, time_step_delta_lst = self._job_dict_lst(self._job_dict)
        return pd.DataFrame({'Job id': job_id_lst,
                             'Start config': time_step_start_lst,
                             'End config': time_step_end_lst,
                             'Step': time_step_delta_lst})

    @property
    def potential_files(self):
        pot = os.path.join(self.working_directory, 'Trained.mtp_')
        states = os.path.join(self.working_directory, 'state.mvs')
        if os.path.exists(pot) and os.path.exists(states):
            return [pot, states]

    def _get_elements(self):
        """
        Return elements in training in insertion order, i.e. elements seen earlier get lower indices.
        """
        elements = []
        for job_id in self._job_dict:
            j = self.project.inspect(job_id)
            if j["NAME"] == "TrainingContainer":
                candidates = j.to_object().get_elements()
            else:
                candidates = j["input/structure/species"]
            for e in candidates:
                if e not in elements:
                    elements.append(e)
        return elements

    def potential_dataframe(self, elements=None):
        """
        :class:`pandas.DataFrame`: potential dataframe for lammps jobs

        .. attention::

            The `elements` argument must be given for in any case for non-unary alloys, as the scrapping code from the
            training data is currently broken!

        Args:
            elements (list of str, optional): name of elements in this potential in the order of their indices in pyiron
                                              structures, if not given use elements seen in training data
        """
        if elements is None:
            elements = self.input["species"]
            if elements is None:
                elements = self._get_elements() # AAAH

        if self.status.finished:
            return pd.DataFrame({
                        "Name": ["".join(elements)],
                        "Filename": [self.potential_files],
                        "Model": ["Custom"],
                        "Species": [elements],
                        "Config": [["pair_style mlip mlip.ini\n",
                                    "pair_coeff * *\n"
                        ]]
            })
        else:
            raise ValueError(f"Potential only available after job is finished, not {self.status}!")

    @property
    def potential(self):
        if self.status.finished and self._potential is not None:
            return self._potential
        else:
            raise ValueError("potential only available on successfully finished jobs")

    def set_input_to_read_only(self):
        """
        This function enforces read-only mode for the input classes, but it has to be implement in the individual
        classes.
        """
        self.input.read_only = True
        self._command_line.read_only = True

    def add_job_to_fitting(self, job_id, time_step_start=0, time_step_end=-1, time_step_delta=10):
        if time_step_end == -1:
            time_step_end = np.shape(self.project.inspect(int(job_id))['output/generic/cells'])[0]-1
        self._job_dict[job_id] = {'time_step_start': time_step_start,
                                  'time_step_end': time_step_end,
                                  'time_step_delta': time_step_delta}

    def write_input(self):
        restart_file_lst = [
                os.path.basename(f)
                for f in self._restart_file_list
            ] + list(self._restart_file_dict.values())
        if 'testing.cfg' not in restart_file_lst:
            species = self._write_test_set(file_name='testing.cfg', cwd=self.working_directory)
            species_count = len(species)
            self.input['species'] = species
        elif self.input['filepath'] != 'auto':
            species_count = 0
        else: 
            raise ValueError('speciescount not set')
        if 'training.cfg' not in restart_file_lst:
            self._write_training_set(file_name='training.cfg', cwd=self.working_directory)
        self._copy_potential(species_count=species_count, file_name='start.mtp', cwd=self.working_directory)
        if self.version != '1.0.0':
            self._command_line.activate_postprocessing()
        self._command_line[0] = self._command_line[0].replace('energy_auto', str(self.input['energy-weight']))
        self._command_line[0] = self._command_line[0].replace('force_auto', str(self.input['force-weight']))
        self._command_line[0] = self._command_line[0].replace('stress_auto', str(self.input['stress-weight']))
        self._command_line[0] = self._command_line[0].replace('iteration_auto', str(int(self.input['iteration'])))
        self._command_line.write_file(file_name='mlip.sh', cwd=self.working_directory)

        species = np.array(species)
        input_store = StructureStorage()
        input_store.add_array('energy', dtype=np.float64, shape=(), per='chunk')
        input_store.add_array('forces', dtype=np.float64, shape=(3,), per='element')
        input_store.add_array('stress', dtype=np.float64, shape=(6,), per='chunk')
        for cfg in loadcfgs(os.path.join(self.working_directory, "training.cfg")):
            struct = Atoms(symbols=species[np.cast[np.int64](cfg.types)], positions=cfg.pos, cell=cfg.lat, pbc=[True]*3) # HACK for pbc
            input_store.add_structure(struct, identifier=cfg.desc,
                                      energy=cfg.energy, forces=cfg.forces, stress=cfg.stresses
            )

        with self.project_hdf5.open('input') as hdf5_input:
            input_store.to_hdf(hdf=hdf5_input, group_name="training_data")

    def collect_logfiles(self):
        pass

    def collect_output(self):
        file_name = os.path.join(self.working_directory, 'diff.cfg')
        if os.path.exists(file_name):
            _, _, _, _, _, _, _, job_id_diff_lst, timestep_diff_lst = read_cgfs(file_name)
        else:
            job_id_diff_lst, timestep_diff_lst = [], []
        file_name = os.path.join(self.working_directory, 'selected.cfg')
        if os.path.exists(file_name):
            _, _, _, _, _, _, _, job_id_new_training_lst, timestep_new_training_lst = read_cgfs(file_name)
        else:
            job_id_new_training_lst, timestep_new_training_lst = [], []
        file_name = os.path.join(self.working_directory, 'grades.cfg')
        if os.path.exists(file_name):
            _, _, _, _, _, _, grades_lst, job_id_grades_lst, timestep_grades_lst = read_cgfs(file_name)
        else:
            grades_lst, job_id_grades_lst, timestep_grades_lst = [], [], []
        try:
            self._potential.load(os.path.join(self.working_directory, "Trained.mtp_"))
        except:
            self.logger.warn('Failed to parse potential file! job.potential will not be available.')
            self._potential = None

        training_store = FlattenedStorage()
        training_store.add_array('energy', dtype=np.float64, shape=(), per='chunk')
        training_store.add_array('forces', dtype=np.float64, shape=(3,), per='element')
        training_store.add_array('stress', dtype=np.float64, shape=(6,), per='chunk')
        for cfg in loadcfgs(os.path.join(self.working_directory, "training_efs.cfg")):
            training_store.add_chunk(len(cfg.pos), identifier=cfg.desc,
                    energy=cfg.energy, forces=cfg.forces, stress=cfg.stresses
            )

        testing_store = FlattenedStorage()
        testing_store.add_array('energy', dtype=np.float64, shape=(), per='chunk')
        testing_store.add_array('forces', dtype=np.float64, shape=(3,), per='element')
        testing_store.add_array('stress', dtype=np.float64, shape=(6,), per='chunk')
        for cfg in loadcfgs(os.path.join(self.working_directory, "testing_efs.cfg")):
            testing_store.add_chunk(len(cfg.pos), identifier=cfg.desc,
                    energy=cfg.energy, forces=cfg.forces, stress=cfg.stresses
            )

        with self.project_hdf5.open('output') as hdf5_output:
            hdf5_output['grades'] = grades_lst
            hdf5_output['job_id'] = job_id_grades_lst
            hdf5_output['timestep'] = timestep_grades_lst
            hdf5_output['job_id_diff'] = job_id_diff_lst
            hdf5_output['timestep_diff'] = timestep_diff_lst
            hdf5_output['job_id_new'] = job_id_new_training_lst
            hdf5_output['timestep_new'] = timestep_new_training_lst
            if self._potential is not None:
                self._potential.to_hdf(hdf=hdf5_output)
            training_store.to_hdf(hdf=hdf5_output, group_name="training_efs")
            testing_store.to_hdf(hdf=hdf5_output, group_name="testing_efs")

    def get_structure(self, iteration_step=-1):
        job = self.project.load(self['output/job_id_diff'][iteration_step])
        return job.get_structure(self['output/timestep_diff'][iteration_step])

    def to_hdf(self, hdf=None, group_name=None):
        super(Mlip, self).to_hdf(hdf=hdf, group_name=group_name)
        self.input.to_hdf(self._hdf5)
        with self._hdf5.open('input') as hdf_input:
            job_id_lst, time_step_start_lst, time_step_end_lst, time_step_delta_lst = self._job_dict_lst(self._job_dict)
            hdf_input["job_id_list"] = job_id_lst
            hdf_input["time_step_start_lst"] = time_step_start_lst
            hdf_input["time_step_end_lst"] = time_step_end_lst
            hdf_input["time_step_delta_lst"] = time_step_delta_lst

    def from_hdf(self, hdf=None, group_name=None):
        super(Mlip, self).from_hdf(hdf=hdf, group_name=group_name)
        self.input.from_hdf(self._hdf5)
        with self._hdf5.open('input') as hdf_input:
            for job_id, start, end, delta in zip(
                    hdf_input["job_id_list"],
                    hdf_input["time_step_start_lst"],
                    hdf_input["time_step_end_lst"],
                    hdf_input["time_step_delta_lst"]
            ):
                self.add_job_to_fitting(
                    job_id=job_id,
                    time_step_start=start,
                    time_step_end=end,
                    time_step_delta=delta
                )
        if self.status.finished:
            with self._hdf5.open("output") as hdf_output:
                if "potential" in hdf_output.list_groups():
                    self._potential.from_hdf(hdf=hdf_output)

    def get_suggested_number_of_configuration(self, species_count=None, multiplication_factor=2.0):
        if self.input['filepath'] == 'auto':
            potential = self.input['potential']
            if isinstance(potential, numbers.Integral):
                potential = f"{potential:02}g.mtp"
            source = self._find_potential(potential)
        else:
            source = self.input['filepath']
        with open(source) as f:
            lines = f.readlines()
        radial_basis_size, radial_funcs_count, alpha_scalar_moments = 0.0, 0.0, 0.0
        for line in lines:
            if 'radial_basis_size' in line:
                radial_basis_size = int(line.split()[2])
            if 'radial_funcs_count' in line:
                radial_funcs_count = int(line.split()[2])
            if 'species_count' in line and species_count is None:
                species_count = int(line.split()[2])
            if 'alpha_scalar_moments' in line:
                alpha_scalar_moments = int(line.split()[2])
        return int(multiplication_factor * \
                   (radial_basis_size * radial_funcs_count * species_count ** 2 + alpha_scalar_moments))

    @staticmethod
    def _update_species_count(lines, species_count):
        ind = 0
        for ind, line in enumerate(lines):
            if 'species_count' in line:
                break
        species_line = lines[ind].split()
        species_line[-1] = str(species_count)
        lines[ind] = ' '.join(species_line) + '\n'
        return lines

    @staticmethod
    def _update_min_max_distance(lines, min_distance, max_distance):
        min_dis_updated, max_dis_updated = False, False
        for ind, line in enumerate(lines):
            if 'min_dist' in line and min_distance is not None:
                min_dist_line = lines[ind].split()
                min_dist_line[2] = str(min_distance)
                lines[ind] = ' '.join(min_dist_line) + '\n'
                min_dis_updated = True
            elif 'max_dist' in line and max_distance is not None:
                max_dist_line = lines[ind].split()
                max_dist_line[2] = str(max_distance)
                lines[ind] = ' '.join(max_dist_line) + '\n'
                max_dis_updated = True
            elif min_dis_updated and max_dis_updated:
                break
        return lines

    def _copy_potential(self, species_count, file_name, cwd=None):
        if cwd is not None:
            file_name = posixpath.join(cwd, file_name)
        if self.input['filepath'] == 'auto':
            potential = self.input['potential']
            if isinstance(potential, numbers.Integral):
                potential = f"{potential:02}g.mtp"
            source = self._find_potential(potential)
        else:
            source = self.input['filepath']
        with open(source) as f:
            lines = f.readlines()
        if self.input['filepath'] == 'auto':
            lines_modified = self._update_species_count(lines, species_count)
            lines_modified = self._update_min_max_distance(lines=lines_modified,
                                                           min_distance=self.input['min_dist'],
                                                           max_distance=self.input['max_dist'])
        else: 
            lines_modified = lines
        with open(file_name, 'w') as f:
            f.writelines(lines_modified)

    @staticmethod
    def stress_tensor_components(stresses):
        return np.array([stresses[0][0], stresses[1][1], stresses[2][2],
                         stresses[1][2], stresses[0][1], stresses[0][2]]) * gpa_to_ev_ang

    def _write_configurations(self, file_name='training.cfg', cwd=None, respect_step=True):
        if cwd is not None:
            file_name = posixpath.join(cwd, file_name)
        indices_lst, position_lst, forces_lst, cell_lst, energy_lst, track_lst, stress_lst = [], [], [], [], [], [], []
        all_species = set()
        for job_id, value in self._job_dict.items():
            ham = self.project.inspect(job_id)
            if respect_step:
                start = value['time_step_start']
                end = value['time_step_end']+1
                delta = value['time_step_delta']
            else:
                start, end, delta = 0, None, 1
            time_step = start
            # HACK: until the training container has a proper HDF5 interface
            if ham.__name__ == "TrainingContainer":
                job = ham.to_object()
                all_species.update(job.get_elements())
                symbol_map = {s: i for i, s in enumerate(sorted(set(all_species)))}
                if end is None:
                    end = job.number_of_structures
                for time_step in range(start, end, delta):
                    symbols = job._container.get_array("symbols", time_step)
                    indices_lst.append([symbol_map[s] for s in symbols])
                    position_lst.append(job._container.get_array("positions", time_step))
                    forces_lst.append(job._container.get_array("forces", time_step))
                    cell_lst.append(job._container.get_array("cell", time_step))
                    energy_lst.append(job._container.get_array("energy", time_step))
                    if job._container.has_array("stress"):
                        volume = np.abs(np.linalg.det(cell_lst[-1]))
                        stress_lst.append(job._container.get_array("stress", time_step) * volume)
                        if np.isnan(stress_lst[-1]).any():
                            stress_lst[-1] = None
                    track_lst.append(str(ham.job_id) + '_' + str(time_step))
                continue
            original_dict = {el: ind for ind, el in enumerate(sorted(ham['input/structure/species']))}
            species_dict = {ind: original_dict[el] for ind, el in enumerate(ham['input/structure/species'])}
            # HACK: this does not preserve element order, but the old code is broken for structures with different
            # species anyway
            all_species.update(ham['input/structure/species'])
            if ham.__name__ in ['Vasp', 'ThermoIntDftEam', 'ThermoIntDftMtp', 'ThermoIntVasp']:
                if len(ham['output/outcar/stresses']) != 0:
                    for position, forces, cell, energy, stresses, volume in zip(ham['output/generic/positions'][start:end:delta],
                                                                                ham['output/generic/forces'][start:end:delta],
                                                                                ham['output/generic/cells'][start:end:delta],
                                                                                ham['output/generic/dft/energy_free'][start:end:delta],
                                                                                ham['output/outcar/stresses'][start:end:delta],
                                                                                ham['output/generic/volume'][start:end:delta]):
                        indices_lst.append([species_dict[el] for el in ham['input/structure/indices']])
                        position_lst.append(position)
                        forces_lst.append(forces)
                        cell_lst.append(cell)
                        energy_lst.append(energy)
                        stress_lst.append(stresses * volume / gpa_to_ev_ang)
                        track_lst.append(str(ham.job_id) + '_' + str(time_step))
                        time_step += delta
                else:
                    for position, forces, cell, energy in zip(ham['output/generic/positions'][start:end:delta],
                                                              ham['output/generic/forces'][start:end:delta],
                                                              ham['output/generic/cells'][start:end:delta],
                                                              ham['output/generic/dft/energy_free'][start:end:delta]):
                        indices_lst.append([species_dict[el] for el in ham['input/structure/indices']])
                        position_lst.append(position)
                        forces_lst.append(forces)
                        cell_lst.append(cell)
                        energy_lst.append(energy)
                        track_lst.append(str(ham.job_id) + '_' + str(time_step))
                        time_step += delta
            elif ham.__name__ in ['Lammps', 'LammpsInt2', 'LammpsMlip']:
                for position, forces, cell, energy, stresses, volume in zip(ham['output/generic/positions'][start:end:delta],
                                                                            ham['output/generic/forces'][start:end:delta],
                                                                            ham['output/generic/cells'][start:end:delta],
                                                                            ham['output/generic/energy_pot'][start:end:delta],
                                                                            ham['output/generic/pressures'][start:end:delta],
                                                                            ham['output/generic/volume'][start:end:delta]):
                    indices_lst.append([species_dict[el] for el in ham['input/structure/indices']])
                    position_lst.append(position)
                    forces_lst.append(forces)
                    cell_lst.append(cell)
                    energy_lst.append(energy)
                    stress_lst.append(self.stress_tensor_components(stresses * volume))
                    track_lst.append(str(ham.job_id) + '_' + str(time_step))
                    time_step += delta
            else:
                for position, forces, cell, energy, stresses, volume in zip(ham['output/generic/positions'][start:end:delta],
                                                                            ham['output/generic/forces'][start:end:delta],
                                                                            ham['output/generic/cells'][start:end:delta],
                                                                            ham['output/generic/energy_pot'][start:end:delta],
                                                                            ham['output/generic/pressures'][start:end:delta],
                                                                            ham['output/generic/volume'][start:end:delta]):
                    indices_lst.append([species_dict[el] for el in ham['input/structure/indices']])
                    position_lst.append(position)
                    forces_lst.append(forces)
                    cell_lst.append(cell)
                    energy_lst.append(energy)
                    stress_lst.append(self.stress_tensor_components(stresses * volume))
                    track_lst.append(str(ham.job_id) + '_' + str(time_step))
                    time_step += delta
        write_cfg(file_name=file_name,
                  indices_lst=indices_lst,
                  position_lst=position_lst,
                  cell_lst=cell_lst,
                  forces_lst=forces_lst,
                  energy_lst=energy_lst,
                  track_lst=track_lst,
                  stress_lst=stress_lst)
        return list(all_species)

    def _write_test_set(self, file_name='testing.cfg', cwd=None):
        return self._write_configurations(file_name=file_name, cwd=cwd, respect_step=False)

    def _write_training_set(self, file_name='training.cfg', cwd=None):
        self._write_configurations(file_name=file_name, cwd=cwd, respect_step=True)


    @staticmethod
    def _job_dict_lst(job_dict):
        job_id_lst, time_step_start_lst, time_step_end_lst, time_step_delta_lst = [], [], [], []
        for key, value in job_dict.items():
            job_id_lst.append(key)
            time_step_start_lst.append(value['time_step_start'])
            time_step_end_lst.append(value['time_step_end'])
            time_step_delta_lst.append(value['time_step_delta'])
        return job_id_lst, time_step_start_lst, time_step_end_lst, time_step_delta_lst

    @staticmethod
    def _find_potential(potential_name):
        for resource_path in state.settings.resource_paths:
            if os.path.exists(os.path.join(resource_path, 'mlip', 'potentials', 'templates')):
                resource_path = os.path.join(resource_path, 'mlip', 'potentials', 'templates')
            if 'potentials' in resource_path and \
                    os.path.exists(os.path.join(resource_path, potential_name)):
                return os.path.join(resource_path, potential_name)
        raise ValueError('Potential not found!')


    # PotentialFit Implementation
    def _add_training_data(self, container):
        self.add_job_to_fitting(container.id, 0, container.number_of_structures - 1, 1)

    def _get_training_data(self):
        # TODO/BUG: only works after input is written for now, instead this should go over _job_
        return self["input/training_data"].to_object()

    def _get_predicted_data(self):
        return self["output/training_efs"].to_object()

    def get_lammps_potential(self):
        return self.potential_dataframe()


class MlipParameter(GenericParameters):
    def __init__(self, separator_char=' ', comment_char='#', table_name="mlip_inp"):
        super(MlipParameter, self).__init__(separator_char=separator_char, comment_char=comment_char,
                                            table_name=table_name)

    def load_default(self, file_content=None):
        if file_content is None:
            file_content = '''\
potential 16g.mtp
filepath auto
energy-weight 1
force-weight 0.01
stress-weight 0.001
iteration 1000
min_dist 2.0
max_dist 5.0
'''
        self.load_string(file_content)


class CommandLine(GenericParameters):
    def __init__(self, input_file_name=None, **qwargs):
        super(CommandLine, self).__init__(input_file_name=input_file_name,
                                          table_name="cmd",
                                          comment_char="#",
                                          val_only=True)

    def load_default(self, file_content=None):
        if file_content is None:
            file_content = '''\
$MLP_COMMAND_PARALLEL train --energy-weight=energy_auto --force-weight=force_auto --stress-weight=stress_auto --max-iter=iteration_auto start.mtp training.cfg > training.log
'''
        self.load_string(file_content)

    def activate_postprocessing(self, file_content=None):
        if file_content is None:
            file_content = '''\
$MLP_COMMAND_PARALLEL train --energy-weight=energy_auto --force-weight=force_auto --stress-weight=stress_auto --max-iter=iteration_auto start.mtp training.cfg > training.log
$MLP_COMMAND_SERIAL calc-grade Trained.mtp_ training.cfg testing.cfg grades.cfg --mvs-filename=state.mvs > grading.log
$MLP_COMMAND_SERIAL calc-errors Trained.mtp_ training.cfg > training.errors
$MLP_COMMAND_SERIAL calc-errors Trained.mtp_ testing.cfg > testing.errors
$MLP_COMMAND_SERIAL calc-efs Trained.mtp_ training.cfg training_efs.cfg
$MLP_COMMAND_SERIAL calc-efs Trained.mtp_ testing.cfg testing_efs.cfg
$MLP_COMMAND_SERIAL select-add Trained.mtp_ training.cfg testing.cfg diff.cfg > select.log
cp training.cfg training_new.cfg 
cat diff.cfg >> training_new.cfg
'''
        self.load_string(file_content)


def write_cfg(file_name, indices_lst, position_lst, cell_lst, forces_lst=None, energy_lst=None, track_lst=None,
              stress_lst=None):
    if stress_lst is None or len(stress_lst) == 0:
        stress_lst = [None] * len(position_lst)
    if forces_lst is None or len(forces_lst) == 0:
        forces_lst = [None] * len(position_lst)
    if track_lst is None or len(track_lst) == 0:
        track_lst = [None] * len(position_lst)
    if energy_lst is None or len(energy_lst) == 0:
        energy_lst = [None] * len(position_lst)
    cfg_lst = []
    for indices, position, forces, cell, energy, stress, track_str in zip(
            indices_lst,
            position_lst,
            forces_lst,
            cell_lst,
            energy_lst,
            stress_lst,
            track_lst
    ):
        cfg_object = Cfg()
        cfg_object.pos = position
        cfg_object.lat = cell
        cfg_object.types = indices
        cfg_object.energy = energy
        cfg_object.forces = forces
        cfg_object.stresses = stress
        if track_str is not None:
            cfg_object.desc = 'pyiron\t' + track_str
        cfg_lst.append(cfg_object)
    savecfgs(filename=file_name, cfgs=cfg_lst, desc=None)


def read_cgfs(file_name):
    cgfs_lst = loadcfgs(filename=file_name, max_cfgs=None)
    cell, positions, forces, stress, energy, indicies, grades, jobids, timesteps = [], [], [], [], [], [], [], [], []
    for cgf in cgfs_lst:
        if cgf.pos is not None:
            positions.append(cgf.pos)
        if cgf.lat is not None:
            cell.append(cgf.lat)
        if cgf.types is not None:
            indicies.append(cgf.types)
        if cgf.energy is not None:
            energy.append(cgf.energy)
        if cgf.forces is not None:
            forces.append(cgf.forces)
        if cgf.stresses is not None:
            stress.append([
                [cgf.stresses[0], cgf.stresses[5], cgf.stresses[4]],
                [cgf.stresses[5], cgf.stresses[1], cgf.stresses[3]],
                [cgf.stresses[4], cgf.stresses[3], cgf.stresses[2]]
            ])
        if cgf.grade is not None:
            grades.append(cgf.grade)
        if cgf.desc is not None:
            job_id, timestep = cgf.desc.split('_')
            jobids.append(int(job_id))
            timesteps.append(int(timestep))
    return [np.array(cell), np.array(positions), np.array(forces), np.array(stress), np.array(energy),
            np.array(indicies), np.array(grades), np.array(jobids), np.array(timesteps)]



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function

import pandas
from pyiron.atomistics.master.parallel import AtomisticParallelMaster
from pyiron_base import JobGenerator

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - " \
                "Computational Materials Design (CM) Department"
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


class ParameterJobGenerator(JobGenerator):
    """
    Job Generator for the ParameterMaster
    """

    @property
    def parameter_list(self):
        """
        List of parameters - a combination of energy cut off and kpoint mesh pairs

        Returns:
            list: of pairs of energy cut off and kpoint mesh
        """
        return [[encut, kpoint_mesh] for encut, kpoint_mesh in zip(
            self._job.iteration_frame.ENCUT,
            self._job.iteration_frame.KPOINT_MESH
        )]

    @staticmethod
    def job_name(parameter):
        """
        Generate job name for a given job based on energy cut off and kpoint mesh
        Args:
            parameter (list): pair of energy cut off and kpoint mesh

        Returns:
            str: job name
        """
        return "job_encut_" + str(parameter[0]).replace('.', '_') + \
               '_kpoints_' + str(parameter[1][0]) + '_' + str(parameter[1][1]) + '_' + str(parameter[1][2])

    def modify_job(self, job, parameter):
        """
        Modify a given job by setting the energy cut off and the kpoint mesh

        Args:
            job (pyiron.dft.job.generic.GenericDFTJob): DFT quantum engine
            parameter (list): pair of energy cut off and kpoint mesh

        Returns:
            pyiron.dft.job.generic.GenericDFTJob: updated job object
        """
        job.set_encut(parameter[0])
        job.set_kpoints(parameter[1])
        return job


class ParameterMaster(AtomisticParallelMaster):
    def __init__(self, project, job_name):
        """
        Parameter master to calculate energies and forces for a given list of energy cut off and kpoint mesh pairs.
        The parameter master is typically used to validate the convergence of a DFT calculation

        Args:
            project (ProjectHDFio): ProjectHDFio instance which points to the HDF5 file the job is stored in
            job_name (str): name of the job, which has to be unique within the project
        """
        super(ParameterMaster, self).__init__(project, job_name)
        self.__name__ = 'ParameterMaster'
        self.__version__ = '0.0.1'
        self._job_generator = ParameterJobGenerator(self)
        self.iteration_frame = pandas.DataFrame({'ENCUT': [], 'KPOINT_MESH': []})

    def to_hdf(self, hdf=None, group_name=None):
        """
        Store the ParameterMaster in an HDF5 file

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(ParameterMaster, self).to_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open('input') as hdf5_input:
            hdf5_input['dataframe'] = self.iteration_frame.to_dict(orient='list')

    def from_hdf(self, hdf=None, group_name=None):
        """
        Restore the ParameterMaster from an HDF5 file

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(ParameterMaster, self).from_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open('input') as hdf5_input:
            self.iteration_frame = pandas.DataFrame(hdf5_input['dataframe'])

    def collect_output(self):
        """
        The ParameterMaster does not store any specitic output, it is recommended to use a pyiron table instead.
        """
        pass



import pandas
from pympipool import Pool
from pyiron_base import DataContainer, GenericJob, state
from pyiron_atomistics.project import Project
from pyiron_atomistics.atomistics.structure.atoms import (
    Atoms,
    pyiron_to_ase,
    ase_to_pyiron,
)
from pyiron_atomistics.atomistics.job.atomistic import AtomisticGenericJob


class GenericJobNoFiles(GenericJob):
    def __init__(self, project, job_name):
        if not state.database.database_is_disabled:
            raise RuntimeError(
                "To run a `Without` job, the database must first be disabled. Please "
                "`from pyiron_base import state; "
                "state.update({'disable_database': True})`, and try again."
            )
        super(GenericJobNoFiles, self).__init__(project, job_name)

        # internal variables
        self._python_only_job = True
        self._interactive_disable_log_file = True

    def refresh_job_status(self):
        if not self._interactive_disable_log_file:
            super(GenericJobNoFiles, self).refresh_job_status()

    def to_hdf(self, hdf=None, group_name=None):
        """

        Args:
            hdf:
            group_name:

        Returns:

        """
        if not self._interactive_disable_log_file:
            super(GenericJobNoFiles, self).to_hdf(hdf=hdf, group_name=group_name)


class AtomisticGenericJobNoFiles(AtomisticGenericJob):
    def __init__(self, project, job_name):
        super(AtomisticGenericJobNoFiles, self).__init__(project, job_name)

        # internal variables
        self._python_only_job = True
        self._interactive_disable_log_file = False

    def refresh_job_status(self):
        if not self._interactive_disable_log_file:
            super(AtomisticGenericJobNoFiles, self).refresh_job_status()

    def to_hdf(self, hdf=None, group_name=None):
        """

        Args:
            hdf:
            group_name:

        Returns:

        """
        if not self._interactive_disable_log_file:
            super(AtomisticGenericJobNoFiles, self).to_hdf(
                hdf=hdf, group_name=group_name
            )
            self._structure_to_hdf()


class AtomisticStructureMasterNoFiles(AtomisticGenericJobNoFiles):
    def __init__(self, project, job_name):
        super(AtomisticStructureMasterNoFiles, self).__init__(project, job_name)
        self._lst_of_struct = []

    @property
    def list_of_structures(self):
        return self._lst_of_struct

    def from_hdf(self, hdf=None, group_name=None):
        super(AtomisticStructureMasterNoFiles, self).from_hdf(
            hdf=hdf, group_name=group_name
        )
        self._structure_from_hdf()
        with self.project_hdf5.open("output/structures") as hdf5_output:
            structure_names = hdf5_output.list_groups()
        for group in structure_names:
            with self.project_hdf5.open("output/structures/" + group) as hdf5_output:
                self._lst_of_struct.append(Atoms().from_hdf(hdf5_output))


class GenericStructureMasterNoFiles(GenericJobNoFiles):
    def __init__(self, project, job_name):
        super(GenericStructureMasterNoFiles, self).__init__(project, job_name)
        self._lst_of_struct = []

    @property
    def list_of_structures(self):
        return self._lst_of_struct

    def from_hdf(self, hdf=None, group_name=None):
        super(GenericStructureMasterNoFiles, self).from_hdf(
            hdf=hdf, group_name=group_name
        )
        self._structure_from_hdf()
        with self.project_hdf5.open("output/structures") as hdf5_output:
            structure_names = hdf5_output.list_groups()
        for group in structure_names:
            with self.project_hdf5.open("output/structures/" + group) as hdf5_output:
                self._lst_of_struct.append(Atoms().from_hdf(hdf5_output))


class SQSMasterMPI(AtomisticStructureMasterNoFiles):
    def __init__(self, project, job_name):
        super(SQSMasterMPI, self).__init__(project, job_name)

        # input
        self.input = DataContainer(table_name="custom_dict")
        self.input.mole_fraction_dict_lst = []

    def to_hdf(self, hdf=None, group_name=None):
        """

        Args:
            hdf:
            group_name:

        Returns:

        """
        if not self._interactive_disable_log_file:
            super(SQSMasterMPI, self).to_hdf(hdf=hdf, group_name=group_name)
            with self.project_hdf5.open("input") as h5in:
                self.input.to_hdf(h5in)

    def run_static(self):
        self.project_hdf5.create_working_directory()
        input_para_lst = [
            [
                i,
                mole_fraction_dict,
                pyiron_to_ase(self.structure),
                self.working_directory,
            ]
            for i, mole_fraction_dict in enumerate(self.input.mole_fraction_dict_lst)
        ]
        with Pool(cores=self.server.cores) as p:
            list_of_structures = p.map(
                function=generate_sqs_structures, lst=input_para_lst
            )
            self._lst_of_struct = [ase_to_pyiron(s) for s in list_of_structures]

        if not self._interactive_disable_log_file:
            for i, structure in enumerate(self._lst_of_struct):
                with self.project_hdf5.open(
                    "output/structures/structure_" + str(i)
                ) as h5:
                    structure.to_hdf(h5)
            self.status.finished = True
            self.project.db.item_update(self._runtime(), self.job_id)


class LAMMPSMinimizeMPI(GenericStructureMasterNoFiles):
    def __init__(self, project, job_name):
        super(LAMMPSMinimizeMPI, self).__init__(project, job_name)

        # input
        self.input = DataContainer(table_name="custom_dict")
        self.input.potential = ""
        self._structure_lst = []

    @property
    def structure_lst(self):
        return self._structure_lst

    @structure_lst.setter
    def structure_lst(self, structure_lst):
        self._structure_lst = structure_lst

    def to_hdf(self, hdf=None, group_name=None):
        """

        Args:
            hdf:
            group_name:

        Returns:

        """
        if not self._interactive_disable_log_file:
            super(LAMMPSMinimizeMPI, self).to_hdf(hdf=hdf, group_name=group_name)
            with self.project_hdf5.open("input") as h5in:
                self.input.to_hdf(h5in)
            with self.project_hdf5.open("input/structures") as hdf5_input:
                for ind, struct in enumerate(self.structure_lst):
                    struct.to_hdf(hdf=hdf5_input, group_name="s_" + str(ind))

    def run_static(self):
        self.project_hdf5.create_working_directory()
        input_para_lst = [
            [i, pyiron_to_ase(structure), self.input.potential, self.working_directory]
            for i, structure in enumerate(self._structure_lst)
        ]
        with Pool(cores=self.server.cores) as p:
            list_of_structures = p.map(
                function=minimize_structure_with_lammps, lst=input_para_lst
            )
            self._lst_of_struct = [ase_to_pyiron(s) for s in list_of_structures]

        if not self._interactive_disable_log_file:
            for i, structure in enumerate(self._lst_of_struct):
                with self.project_hdf5.open(
                    "output/structures/structure_" + str(i)
                ) as h5:
                    structure.to_hdf(h5)
            self.status.finished = True
            self.project.db.item_update(self._runtime(), self.job_id)


class LAMMPSElasticMPI(GenericJobNoFiles):
    def __init__(self, project, job_name):
        super(LAMMPSElasticMPI, self).__init__(project, job_name)

        # input
        self.input = DataContainer(table_name="custom_dict")
        self.input.potential = ""
        self.input.element_lst = []
        self._structure_lst = []
        self._results_df = None

    @property
    def structure_lst(self):
        return self._structure_lst

    @structure_lst.setter
    def structure_lst(self, structure_lst):
        self._structure_lst = structure_lst

    @property
    def results(self):
        return self._results_df

    def to_hdf(self, hdf=None, group_name=None):
        """

        Args:
            hdf:
            group_name:

        Returns:

        """
        if not self._interactive_disable_log_file:
            super(LAMMPSElasticMPI, self).to_hdf(hdf=hdf, group_name=group_name)
            with self.project_hdf5.open("input") as h5in:
                self.input.to_hdf(h5in)
            with self.project_hdf5.open("input/structures") as hdf5_input:
                for ind, struct in enumerate(self.structure_lst):
                    struct.to_hdf(hdf=hdf5_input, group_name="s_" + str(ind))

    def run_static(self):
        self.project_hdf5.create_working_directory()
        input_para_lst = [
            [
                i,
                pyiron_to_ase(structure),
                self.input.element_lst,
                self.input.potential,
                self.working_directory,
            ]
            for i, structure in enumerate(self._structure_lst)
        ]
        with Pool(cores=self.server.cores) as p:
            results = p.map(function=get_elastic_constants, lst=input_para_lst)

        self._results_df = convert_elastic_constants_to_dataframe(results)
        if not self._interactive_disable_log_file:
            self._results_df.to_hdf(
                self.project_hdf5._file_name, self.job_name + "/output/df"
            )

    def from_hdf(self, hdf=None, group_name=None):
        super(LAMMPSElasticMPI, self).from_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("output") as hdf5_output:
            if "df" in hdf5_output.list_groups():
                self._results_df = pandas.read_hdf(
                    self.project_hdf5._file_name, self.job_name + "/output/df"
                )


class LAMMPSMinimizeElasticMPI(AtomisticStructureMasterNoFiles):
    def __init__(self, project, job_name):
        super(LAMMPSMinimizeElasticMPI, self).__init__(project, job_name)

        # input
        self.input = DataContainer(table_name="custom_dict")
        self.input.mole_fraction_dict_lst = []
        self.input.potential = ""
        self.input.element_lst = []

    @property
    def results(self):
        return self._results_df

    def to_hdf(self, hdf=None, group_name=None):
        """

        Args:
            hdf:
            group_name:

        Returns:

        """
        if not self._interactive_disable_log_file:
            super(LAMMPSMinimizeElasticMPI, self).to_hdf(hdf=hdf, group_name=group_name)
            with self.project_hdf5.open("input") as h5in:
                self.input.to_hdf(h5in)

    def run_static(self):
        self.project_hdf5.create_working_directory()
        input_para_lst = [
            [
                i,
                mole_fraction_dict,
                pyiron_to_ase(self.structure),
                self.input.element_lst,
                self.input.potential,
                self.working_directory,
            ]
            for i, mole_fraction_dict in enumerate(self.input.mole_fraction_dict_lst)
        ]
        with Pool(cores=self.server.cores) as p:
            results = p.map(function=combined_function, lst=input_para_lst)

        self._results_df = convert_elastic_constants_to_dataframe(results)
        if not self._interactive_disable_log_file:
            self._results_df.to_hdf(
                self.project_hdf5._file_name, self.job_name + "/output/df"
            )

    def from_hdf(self, hdf=None, group_name=None):
        super(LAMMPSMinimizeElasticMPI, self).from_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("output") as hdf5_output:
            if "df" in hdf5_output.list_groups():
                self._results_df = pandas.read_hdf(
                    self.project_hdf5._file_name, self.job_name + "/output/df"
                )


def generate_sqs_structures(input_parameter):
    i, mole_fraction_dict, structure_template, working_directory = input_parameter

    # import
    import numpy as np
    from pyiron_atomistics import Project, ase_to_pyiron, pyiron_to_ase
    from pyiron_contrib.nofiles.sqs import SQSJobWithoutOutput

    # calculation
    if len(mole_fraction_dict) > 1:
        project = Project(working_directory)
        job = project.create_job(SQSJobWithoutOutput, "sqs_" + str(i))
        job._interactive_disable_log_file = True
        job.structure = ase_to_pyiron(structure_template)
        job.input["mole_fractions"] = mole_fraction_dict
        job.input["iterations"] = 1e6
        job.server.cores = 1
        job.run()
        structure_next = pyiron_to_ase(job._lst_of_struct[-1])
    else:
        # use ASE syntax
        structure_next = structure_template.copy()
        structure_next.symbols[:] = list(mole_fraction_dict.keys())[-1]

    # return value
    return structure_next


def minimize_structure_with_lammps(input_parameter):
    i, structure_next, potential, working_directory = input_parameter

    # import
    from pyiron_atomistics import Project, ase_to_pyiron, pyiron_to_ase
    from pyiron_contrib.nofiles.lammps import LammpsInteractiveWithoutOutput
    from mpi4py import MPI

    # calculation
    project = Project(working_directory)
    lmp_mini1 = project.create_job(
        LammpsInteractiveWithoutOutput, "lmp_mini_" + str(i), delete_existing_job=True
    )
    lmp_mini1.structure = ase_to_pyiron(structure_next)
    lmp_mini1.potential = potential
    lmp_mini1.calc_minimize(pressure=0.0)
    lmp_mini1.server.run_mode.interactive = True
    lmp_mini1.interactive_mpi_communicator = MPI.COMM_SELF
    lmp_mini1._interactive_disable_log_file = True  # disable lammps.log
    lmp_mini1.run()
    lmp_mini1.interactive_close()

    # return value
    return pyiron_to_ase(lmp_mini1.get_structure())


def get_elastic_constants(input_para):
    i, structure, element_lst, potential, working_directory = input_para

    # import
    from pyiron_atomistics import Project, ase_to_pyiron
    from pyiron_contrib.nofiles.lammps import LammpsInteractiveWithoutOutput
    from pyiron_contrib.nofiles.elastic import ElasticMatrixJobWithoutFiles
    from mpi4py import MPI

    # Elastic constants
    project = Project(working_directory)
    lmp_elastic = project.create_job(
        LammpsInteractiveWithoutOutput,
        "lmp_elastic_" + str(i),
        delete_existing_job=True,
    )
    lmp_elastic.structure = ase_to_pyiron(structure)
    lmp_elastic.potential = potential
    lmp_elastic.interactive_enforce_structure_reset = True
    lmp_elastic.interactive_mpi_communicator = MPI.COMM_SELF
    lmp_elastic.server.run_mode.interactive = True
    lmp_elastic._interactive_disable_log_file = True  # disable lammps.log
    elastic = lmp_elastic.create_job(
        ElasticMatrixJobWithoutFiles, "elastic_" + str(i), delete_existing_job=True
    )
    elastic._interactive_disable_log_file = True  # disable lammps.log
    elastic.run()

    # return value
    elastic_constants_lst = [
        elastic._data["C"][0][0],
        elastic._data["C"][0][1],
        elastic._data["C"][0][2],
        elastic._data["C"][0][3],
        elastic._data["C"][0][4],
        elastic._data["C"][0][5],
        elastic._data["C"][1][1],
        elastic._data["C"][1][2],
        elastic._data["C"][1][3],
        elastic._data["C"][1][4],
        elastic._data["C"][1][5],
        elastic._data["C"][2][2],
        elastic._data["C"][2][3],
        elastic._data["C"][2][4],
        elastic._data["C"][2][5],
        elastic._data["C"][3][3],
        elastic._data["C"][3][4],
        elastic._data["C"][3][5],
        elastic._data["C"][4][4],
        elastic._data["C"][4][5],
        elastic._data["C"][5][5],
    ]

    conc_lst = []
    for el in element_lst:
        if el in elastic.ref_job.structure.get_species_symbols():
            conc_lst.append(
                sum(
                    elastic.ref_job.structure.indices
                    == elastic.ref_job.structure.get_species_symbols()
                    .tolist()
                    .index(el)
                )
                / len(elastic.ref_job.structure.indices)
            )
        else:
            conc_lst.append(0.0)

    return elastic_constants_lst + conc_lst


def combined_function(input_parameter):
    (
        i,
        mole_fraction_dict,
        structure_template,
        element_lst,
        potential,
        working_directory,
    ) = input_parameter

    # import
    from pyiron_contrib.nofiles.master import (
        generate_sqs_structures,
        minimize_structure_with_lammps,
        get_elastic_constants,
    )

    # calculation
    structure_next = generate_sqs_structures(
        input_parameter=[i, mole_fraction_dict, structure_template, working_directory]
    )
    structure = minimize_structure_with_lammps(
        input_parameter=[i, structure_next, potential, working_directory]
    )
    results = get_elastic_constants(
        input_para=[i, structure, element_lst, potential, working_directory]
    )

    # return value
    return results


def convert_elastic_constants_to_dataframe(results):
    (
        c11_lst,
        c12_lst,
        c13_lst,
        c14_lst,
        c15_lst,
        c16_lst,
        c22_lst,
        c23_lst,
        c24_lst,
        c25_lst,
        c26_lst,
        c33_lst,
        c34_lst,
        c35_lst,
        c36_lst,
        c44_lst,
        c45_lst,
        c46_lst,
        c55_lst,
        c56_lst,
        c66_lst,
        conc_Fe_lst,
        conc_Ni_lst,
        conc_Cr_lst,
        conc_Co_lst,
        conc_Cu_lst,
    ) = zip(*results)

    return pandas.DataFrame(
        {
            "conc_Fe": conc_Fe_lst,
            "conc_Ni": conc_Ni_lst,
            "conc_Cr": conc_Cr_lst,
            "conc_Co": conc_Co_lst,
            "conc_Cu": conc_Cu_lst,
            "C11": c11_lst,
            "C12": c12_lst,
            "C13": c13_lst,
            "C14": c14_lst,
            "C15": c15_lst,
            "C16": c16_lst,
            "C22": c22_lst,
            "C23": c23_lst,
            "C24": c24_lst,
            "C25": c25_lst,
            "C26": c26_lst,
            "C33": c33_lst,
            "C34": c34_lst,
            "C35": c35_lst,
            "C36": c36_lst,
            "C44": c44_lst,
            "C45": c45_lst,
            "C46": c46_lst,
            "C55": c55_lst,
            "C56": c56_lst,
            "C66": c66_lst,
        }
    )




from pyiron_contrib.protocol.utils.misc import LoggerMixin, ensure_iterable, flatten, requires_arguments, \
    ordered_dict_get_index, ordered_dict_get_last
from pyiron_contrib.protocol.utils.dictionaries import IODictionary, InputDictionary, TimelineDict
from pyiron_contrib.protocol.utils.pointer import Pointer, Path, Crumb, CrumbType
from pyiron_contrib.protocol.utils.event import Event, EventHandler
from pyiron_contrib.protocol.utils.comparers import Comparer


# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function
import numpy as np

"""
Simetimes numpy and scipy are missing things, or I can't find them.
"""

__author__ = "Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "June 26, 2019"


def welford_online(x, mean, std, k):
    """
    Computes the cummulative mean and standard deviation.

    Note: The standard deviation calculated is for the population (ddof=0). For the sample (ddof=1) it would need to
    be extended.

    Args:
        x (float/numpy.ndarray): The new sample.
        mean (float/numpy.ndarray): The mean so far.
        std (float/numpy.ndarray): The standard deviation so far.
        k (int): How many samples were used to calculate the existing `mean` and `std`.

    Returns:
        float/numpy.ndarray, float/numpy.ndarray: The new mean and standard deviation for `k+1` values.
    """
    new_mean = (x + k * mean) / (k + 1)
    new_std = np.sqrt((k * std ** 2 + (x - mean) * (x - new_mean)) / (k + 1))
    return new_mean, new_std


"""Generic File Object."""

# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
import posixpath
import shutil

from abc import ABC, abstractmethod
import os
from functools import lru_cache
from pathlib import Path

import pandas as pd

from pyiron_base import GenericJob, state
from pyiron_base import (
    FileDataTemplate as BaseFileDataTemplate,
    load_file,
    FileData as FileDataBase,
)
from pyiron_base import HasGroups

try:
    import ipywidgets as widgets
    from IPython.display import display

    gui_elements_imported = True
except ImportError:
    gui_elements_imported = False

__author__ = "Niklas Siemer"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.1"
__maintainer__ = "Niklas Siemer"
__email__ = "siemer@mpie.de"
__status__ = "development"
__date__ = "Feb 02, 2021"


class MetaDataTemplate(ABC):
    @abstractmethod
    def to_dict(self):
        """Provide a dict representation of the meta data."""

    def _repr_html_(self):
        df = pd.DataFrame(
            {
                "Metadata": list(self.to_dict().keys()),
                "  ": list(self.to_dict().values()),
            }
        )
        df.set_index("Metadata", inplace=True)
        df[df.isnull()] = ""
        return df._repr_html_()


class MetaData(MetaDataTemplate):
    def __init__(self, metadata_dict=None):
        if metadata_dict is None:
            metadata_dict = {}
        self._metadata = {key: val for key, val in metadata_dict.items()}

    def to_dict(self):
        return self._metadata.copy()

    def __repr__(self):
        return f"{self.__class__.__name__}({self._metadata})"


class FileDataTemplate(BaseFileDataTemplate, ABC):
    def __init__(self):
        self._metadata = None
        if gui_elements_imported:
            self._output = widgets.Output()
            self._box = widgets.VBox([self._output])

    @staticmethod
    def _get_filetype_from_filename(filename):
        filetype = Path(filename).suffix
        return None if filetype == "" else filetype[1:]

    def _get_metadata(self):
        return self._metadata

    def _set_metadata(self, metadata):
        if not isinstance(metadata, MetaDataTemplate):
            self._metadata = MetaData(metadata)
        else:
            self._metadata = metadata

    @property
    def metadata(self):
        return self._get_metadata()

    @metadata.setter
    def metadata(self, metadata):
        self._set_metadata(metadata)

    if gui_elements_imported:

        @property
        def gui(self):
            self._output.clear_output()
            with self._output:
                display(self.data, self.metadata)
            return self._box

        def _ipython_display_(self):
            display(self.gui)


class FileData(FileDataTemplate):
    """FileData stores an instance of a data file, e.g. a single Image from a measurement."""

    def __init__(self, file, data=None, metadata=None, filetype=None):
        """FileData class to store data and associated metadata.

        Args:
            file (str): path to the data file (if data is None) or filename associated with the data.
            data (object/None): object containing data
            metadata (dict/DataContainer): Dictionary of metadata associated with the data
            filetype (str): File extension associated with the type data,
                            If provided this overwrites the assumption based on the extension of the filename.
        """
        super().__init__()
        if data is None:
            self.filename = os.path.split(file)[1]
            self.source = file
            self._data = None
        else:
            self.filename = file
            self.source = None
            self._data = data
        if filetype is None:
            self.filetype = self._get_filetype_from_filename(self.filename)
        else:
            self.filetype = filetype
        if metadata is None:
            self.metadata = MetaData()
        elif isinstance(metadata, MetaDataTemplate):
            self.metadata = metadata
        else:
            self.metadata = MetaData(metadata)
        self._hasdata = self._data is not None

    @property
    @lru_cache()
    def data(self):
        """Return the associated data."""
        if self._hasdata:
            return self._data
        else:
            return load_file(self.source, filetype=self.filetype)


class StorageInterface(HasGroups, ABC):
    """File handling in different storage interfaces"""

    def __init__(self):
        self._path = None

    @property
    def path(self):
        """Represents the path of the storage interface (always posix-path)"""
        return self._path

    @path.setter
    def path(self, new_path):
        self._path = self._path_setter(new_path)

    def _path_setter(self, new_path):
        return new_path

    @abstractmethod
    def is_file(self, item):
        """Check if the given item is a node/file"""

    @abstractmethod
    def is_dir(self, item):
        """Check if the given item is a group/directory"""

    @abstractmethod
    def upload_file(self, file, metadata=None, filename=None):
        """Upload the provided files to the storage"""

    @abstractmethod
    def remove_file(self, file):
        """Removes specified files from the storage"""

    @abstractmethod
    def __getitem__(self, item):
        """Return stored file as Subclass of FileDataTemplate"""

    @property
    def requires_metadata(self):
        return False

    def setup_storage(self):
        pass

    def parse_metadata(self, metadata):
        return metadata

    @abstractmethod
    def validate_metadata(self, metadata, raise_error=True):
        """Check metadata for validity and provide valid metadata back.

        Args:
            metadata: the meta data object to check
            raise_error: if raise_error is True, errors are raised. Otherwise, silently returning None.
        Raises:
            ValueError: if the metadata is not valid and raise_error.
        Returns:
            object: valid meta data or None if metadata is not valid and not raise_error.
        """


class LocalFileStorage(StorageInterface):
    def upload_file(self, file, metadata=None, filename=None):
        self.put(file=file, filepath=filename)

    def __init__(self, path, root_path=None):
        super().__init__()
        self.path = path
        if root_path is not None:
            self._root_path = Path(root_path)
        else:
            self._root_path = root_path

    def _path_setter(self, new_path):
        path = Path(new_path).expanduser()
        if not path.is_dir():
            raise ValueError(f"Provided path '{new_path}' is not a directory.")
        return path.resolve()

    def create_group(self, name):
        new_path = self._join_path(name)
        if new_path.exists() and not new_path.is_dir():
            raise ValueError(f"{new_path} already exists and is not a directory.")
        elif not new_path.exists():
            new_path.mkdir(parents=True)

        return self.__class__(new_path)

    def is_file(self, item):
        return self._join_path(item).is_file()

    def is_dir(self, item):
        return self._join_path(item).is_dir()

    def _list_groups(self):
        return [g for g in os.listdir(self.path) if self.is_dir(g)]

    def _list_nodes(self):
        return [f for f in os.listdir(self.path) if self.is_file(f)]

    def _join_path(self, item):
        if os.path.isabs(item):
            return item
        return self._path.joinpath(item).resolve()

    def __getitem__(self, item):
        item_path = self._join_path(item)
        if item_path.is_file():
            return FileData(str(item_path))
        elif item_path.is_dir():
            return self.__class__(item_path)
        else:
            raise KeyError(f"No such file or directory: '{item}' in '{self._path}'")

    def validate_metadata(self, metadata, raise_error=True):
        state.logger.warn(
            "Storing metadata for LocalStorage is currently handled only on the job level."
        )
        return metadata

    def put(self, file, filepath=None, overwrite=False):
        if filepath is None:
            filepath = self._path
        else:
            filepath = self._join_path(filepath)

        if filepath.is_file() and not overwrite:
            raise ValueError(f"File already present and overwrite=False.")

        if isinstance(file, (str, Path)):
            shutil.copy2(file, filepath)
        else:
            # Expecting file to be a file handle
            if filepath.is_dir():
                filepath = filepath.joinpath(file.name)
            shutil.copyfileobj(file, filepath.open("wb"))

    def remove_file(self, filename, missing_ok=False):
        file = self._join_path(filename)
        if file.is_dir():
            raise ValueError(f"{filename} is a directory but expected a file.")
        file.unlink(missing_ok=missing_ok)


class JobFileStorage(LocalFileStorage):
    """The job storage operates on the usual working directory of the job"""

    def _path_setter(self, new_path):
        new_path = super()._path_setter(new_path)
        self._validate_path(new_path)
        self._path = new_path

    def __init__(self, working_directory):
        super().__init__(path=".", root_path=working_directory)

    def _join_path(self, item):
        new_path = super()._join_path(item)
        self._validate_path(new_path)
        return new_path

    def _validate_path(self, path):
        rel_path = os.path.relpath(path, self._root_path)
        if rel_path.startswith(".."):
            raise ValueError(
                f"New path not within the jobs working directory {self._root_path}"
            )

    def setup_storage(self):
        os.mkdir(self._root_path)



import xml.etree.ElementTree as ET

from pyiron_base import DataContainer


class Constraint(DataContainer):
    def __init__(self, identifier=None, dependent_dof=None, expression=None) -> None:
        super().__init__(table_name=f"constraint_{identifier}")
        self.identifier = identifier
        self.dependent_dof = dependent_dof
        self.expression = expression

    def _to_xml_element(self):
        root = ET.Element("constraint")
        root.set("id", self.identifier)
        root.set("dependent-dof", self.dependent_dof)
        expr = ET.SubElement(root, "expression")
        expr.text = self.expression
        return root


class ParameterConstraints(DataContainer):
    def __init__(self, table_name="parameter_constraints") -> None:
        super().__init__(table_name=table_name)

    def add_constraint(self, identifier, dependent_dof, expression):
        self[identifier] = Constraint(
            identifier=identifier, dependent_dof=dependent_dof, expression=expression
        )

    def _to_xml_element(self):
        root = ET.Element("parameter-constraints")
        for c in self.values():
            root.append(c._to_xml_element())
        return root



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""Pyiron interface to atomicrex"""
import numpy as np
import pandas as pd

from pyiron_base import state, GenericJob, Executable, FlattenedStorage

from pyiron_contrib.atomistics.atomicrex.general_input import (
    GeneralARInput,
    AlgorithmFactory,
)
from pyiron_contrib.atomistics.atomicrex.structure_list import ARStructureContainer
from pyiron_contrib.atomistics.atomicrex.potential_factory import ARPotFactory
from pyiron_contrib.atomistics.atomicrex.output import Output
from pyiron_contrib.atomistics.atomicrex.function_factory import FunctionFactory
from pyiron_contrib.atomistics.ml.potentialfit import PotentialFit
from pyiron_contrib.atomistics.atomistics.job.trainingcontainer import (
    TrainingContainer,
    TrainingStorage,
)


class AtomicrexBase(GenericJob, PotentialFit):
    __version__ = "0.1.0"
    __hdf_version__ = "0.1.0"
    """Class to set up and run atomicrex jobs"""

    def __init__(self, project, job_name):
        super().__init__(project, job_name)
        # self._executable_activate(enforce=True)
        state.publications.add(self.publication)
        self.input = GeneralARInput()
        self.potential = None
        self.structures = ARStructureContainer()
        self.output = Output()
        self.factories = Factories()
        self._compress_by_default = True

    def plot_final_potential(self):
        """
        Plot the fitted potential.
        Returns the matplotlib objects to change the look of the plot.

        Returns:
            [matplotlib figure, axis]: [description]
        """
        return self.potential.plot_final_potential(self)

    def to_hdf(self, hdf=None, group_name=None):
        """Internal function to store the job in hdf5 format"""
        super().to_hdf(hdf=hdf, group_name=group_name)
        self.input.to_hdf(hdf=self.project_hdf5)
        self.potential.to_hdf(hdf=self.project_hdf5)
        self.structures.to_hdf(hdf=self.project_hdf5)
        self.output.to_hdf(hdf=self.project_hdf5)

    def from_hdf(self, hdf=None, group_name=None):
        """Internal function to reload the job object from hdf5"""
        super().from_hdf(hdf=hdf, group_name=group_name)
        self.input.from_hdf(hdf=self.project_hdf5)
        self.potential = self.project_hdf5["potential"].to_object()
        self.structures.from_hdf(hdf=self.project_hdf5)
        self.output.from_hdf(hdf=self.project_hdf5)

    @property
    def publication(self):
        return {
            "atomicrex": {
                "atomicrex": {
                    "title": "Atomicrex-a general purpose tool for the construction of atomic interaction models",
                    "journal": "Modelling and Simulation in Materials Science and Engineering",
                    "volume": "25",
                    "number": "5",
                    "year": "2017",
                    "issn": "0965-0393",
                    "doi": "10.1088/1361-651X/aa6ecf",
                    "url": "https://doi.org/10.1088%2F1361-651x%2Faa6ecf",
                    "author": [
                        "Alexander Stukowski",
                        "Erik Fransson",
                        "Markus Mock",
                        "Paul Erhart",
                    ],
                }
            }
        }

    def compress(self, files_to_compress=None):
        """
        Compress the output files of a job object.

        Args:
            files_to_compress (list): A list of files to compress (optional)
        """
        if files_to_compress is None:
            if self.potential.export_file is None:
                files_to_compress = self.list_files()
            else:
                files_to_compress = [
                    f for f in self.list_files() if f != self.potential.export_file
                ]
        super().compress(files_to_compress=files_to_compress)

    def collect_output(self, cwd=None):
        """Internal function that parses the output of an atomicrex job

        Args:
            cwd (str, optional): Working directory. Defaults to None.
        """
        # self.input.from_hdf(self._hdf5)
        if cwd is None:
            cwd = self.working_directory
        if self.input.__version__ == "0.1.0":
            filepath = f"{cwd}/atomicrex.out"
        else:
            filepath = f"{cwd}/error.out"

        finished_triggered = False
        params_triggered = False
        dependent_dofs_triggered = False
        structures_triggered = False

        # Allocate numpy arrays for iterations and residual
        # I assume this is better than appending to a list if many iterations are done
        if self.input.fit_algorithm.name == "BFGS":
            residuals = np.zeros(self.input.fit_algorithm.max_iter + 1)
        else:
            residuals = np.zeros(self.input.fit_algorithm.max_iter)

        # Since every step is written out in atomicrex arange can be used.
        # Needs to be adapted when atomicrex output is changed to write only every xth step.
        # Unsinged 32 bit int should be enough or this will overflow anyway in most cases.
        iterations = np.arange(start=1, stop=len(residuals) + 1, dtype=np.uintc)
        iter_index = 0

        with open(filepath, "r") as f:
            final_parameter_lines = []
            final_property_lines = []
            depdendent_dof_lines = []

            for l in f:
                if l.startswith("ERROR"):
                    self.status.aborted = True
                    self.output.error = l

                elif not finished_triggered:
                    if l.startswith("Iterations"):
                        l = l.split()
                        finished_triggered = True
                        self.output.residual = residuals[0:iter_index]
                        self.output.iterations = iterations[0:iter_index]
                    else:
                        l = l.split()
                        try:
                            if l[1] == "iter=":
                                residuals[iter_index] = float(l[-1])
                                iter_index += 1
                        except IndexError:
                            continue

                else:  # if finished_triggered
                    if params_triggered:
                        if not l.startswith("---"):
                            final_parameter_lines.append(l)
                        else:
                            # Collecting lines with final parameters finished, hand over to the potential class
                            self.potential._parse_final_parameters(
                                final_parameter_lines
                            )
                            params_triggered = False

                    elif structures_triggered:
                        if not l.startswith("---"):
                            final_property_lines.append(l)
                        else:
                            # Collecting structure information finished, hand over structures class
                            self.structures._parse_final_properties(
                                final_property_lines
                            )
                            structures_triggered = False

                    elif dependent_dofs_triggered:
                        if not l.startswith("---"):
                            depdendent_dof_lines.append(l)
                        else:
                            self.potential._parse_final_parameters(depdendent_dof_lines)
                            dependent_dofs_triggered = False

                    elif l.startswith("Potential parameters"):
                        # Get the number of dofs
                        n_fit_dofs = int(l.split("=")[1][:-3])
                        params_triggered = True

                    elif l.startswith("Computing"):
                        structures_triggered = True

                    elif l.startswith("Dependent DOFs:"):
                        dependent_dofs_triggered = True
        self.to_hdf()

    def convergence_check(self):
        """
        Internal function, TODO
        find something to reasonably judge convegence
        """
        return True

    def write_input(self, directory=None):
        """Internal function that writes input files

        Args:
            directory ([string], optional): Working directory. Defaults to None.
        """
        if directory is None:
            directory = self.working_directory
        self.input._write_xml_file(directory=directory, job=self)
        self.potential.write_xml_file(directory=directory)
        self.structures.write_xml_file(directory=directory)

    def _executable_activate(self, enforce=False):
        """
        Internal function that sets up and Executable() object
        and finds executables available in pyiron resources/atomicrex/bin

        Args:
            enforce (bool, optional): [description]. Defaults to False.
        """
        if self._executable is None or enforce:
            if len(self.__module__.split(".")) > 1:
                self._executable = Executable(
                    codename=self.__name__,
                    module=self.__module__.split(".")[-2],
                    path_binary_codes=state.settings.resource_paths,
                )
            else:
                self._executable = Executable(
                    codename=self.__name__,
                    path_binary_codes=state.settings.resource_paths,
                )

    # Leftover of the potentials workshop.
    # Maybe this property will be used in unified interface
    # to several fitting codes in the future
    # instead of the potential_as_pd_df function
    @property
    def lammps_potential(self):
        return self.potential_as_pd_df()

    def potential_as_pd_df(self):
        """
        Return the fitted potential as a pandas dataframe,
        which can be used for lammps calculations.
        """
        return self.potential._potential_as_pd_df(job=self)

    #### PotentialFit methods
    def _add_training_data(self, container: TrainingContainer) -> None:
        self.structures.add_training_data(container)

    def _get_training_data(self) -> TrainingStorage:
        return self.structures.get_training_data()

    def _get_predicted_data(self) -> FlattenedStorage:
        return self.structures.get_predicted_data()

    def get_lammps_potential(self) -> pd.DataFrame:
        """
        Return a pyiron compatible dataframe that defines a potential to be used with a Lammps job (or subclass
        thereof).

        Returns:
            DataFrame: contains potential information to be used with a Lammps job.
        """
        return self.potential_as_pd_df()


class Factories:
    """
    Provides conventient acces to other factory classes.
    Functionality to set up an atomicrex job can be found here.
    """

    def __init__(self):
        self.potentials = ARPotFactory()
        self.functions = FunctionFactory()
        self.algorithms = AlgorithmFactory()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import os
from pyiron_contrib.atomistics.mlip.mlip import Mlip, read_cgfs

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - " \
                "Computational Materials Design (CM) Department"
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"


class MlipSelect(Mlip):
    def __init__(self, project, job_name):
        super(MlipSelect, self).__init__(project, job_name)
        self.__version__ = None
        self.__name__ = "MlipSelect"
        self._executable_activate()
        del self.input['min_dist']
        del self.input['max_dist']
        del self.input['iteration']
        del self.input['energy-weight']
        del self.input['force-weight']
        del self.input['stress-weight']
        self._command_line._delete_line(0)
        self._command_line._delete_line(0)

    def validate_ready_to_run(self):
        if len(self.restart_file_list) == 0:
            raise ValueError()

    def write_input(self):
        species_count = self._write_test_set(file_name='testing.cfg', cwd=self.working_directory)
        self._copy_potential(species_count, file_name='start.mtp', cwd=self.working_directory)
        self._command_line[0] = self._command_line[0].replace('Trained.mtp_', 'start.mtp')
        self._command_line.write_file(file_name='mlip.sh', cwd=self.working_directory)

    def collect_output(self):
        file_name = os.path.join(self.working_directory, 'diff.cfg')
        cell, positions, forces, stress, energy, indicies, grades, jobids, timesteps = read_cgfs(file_name=file_name)
        with self.project_hdf5.open('output') as hdf5_output:
            hdf5_output['forces'] = forces
            hdf5_output['energy_tot'] = energy
            hdf5_output['pressures'] = stress
            hdf5_output['cells'] = cell
            hdf5_output['positions'] = positions
            hdf5_output['indicies'] = indicies
